# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import gzip
from os.path import basename
from metagenomix._io_utils import (io_update, to_do, tech_specificity,
                                   status_update)
from metagenomix.core.parameters import tech_params


def get_cat_zcat(
        fastq_fp: str
) -> str:
    """Get the command `cat` or `gzcat` to run for fastq editing.

    Parameters
    ----------
    fastq_fp : str
        Fastq file path.

    Returns
    -------
    cat : str
        The command to use (`cat` or `zcat`).
    """
    cat = "cat"
    if fastq_fp.endswith('.fastq.gz'):
        cat = "zcat"
    return cat


def edit_fastq_cmd(
        fastq_fp: str,
        num: int
) -> str:
    """Get the unix command to run on each fastq file that
    needs editing, depending on the source of fastq file.

    Parameters
    ----------
    fastq_fp : str
        Fastq file path.
    num : int
        Can be 1 or 2 for the forward and reverse read, respectively.

    Returns
    -------
    cmd : str
        Unix command to run to edit a fastq file.
    """
    cat = get_cat_zcat(fastq_fp)

    cmd = 'bioawk_exe=$(which bioawk)\n'
    cmd += 'if [ -x "$bioawk_exe" ] ; then\n'
    cmd += '%s %s | ' % (cat, fastq_fp)
    cmd += "bioawk -c fastx "
    cmd += "'{print \"@\"$1\"/%s\\n\"$2\"\\n+\\n\"$3}' " % str(num)
    if fastq_fp.endswith('.fastq'):
        cmd += " > %s_renamed\n" % fastq_fp
    else:
        cmd += " | gzip > %s_renamed\n" % fastq_fp
    cmd += "mv %s_renamed %s\n" % (fastq_fp, fastq_fp)
    cmd += 'else\n'
    cmd += 'echo "Need to load bioawk module or environment with bioawk"\n'
    cmd += 'exit 1\n'
    cmd += 'fi\n'
    return cmd


def get_fastq_header(
        fastq_path: str
) -> str:
    """Get the first line of the fastq file.

    Parameters
    ----------
    fastq_path : str
        Path to a fastq.gz or fastq file.

    Returns
    -------
    fastq_line : str
        First line of the fastq file.
    """
    fastq_line = ''
    if fastq_path.endswith('fastq.gz'):
        with gzip.open(fastq_path) as f:
            for line in f:
                fastq_line = line.decode().strip()
                break
    elif fastq_path.endswith('fastq'):
        with open(fastq_path) as f:
            for line in f:
                fastq_line = line.strip()
                break
    return fastq_line


def get_edit_cmd(
        self,
        num: int,
        tech: str
) -> None:
    """Get the unix command to run on each fastq file that needs editing.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .config
            Configurations
    num : int
        Can be 1 or 2 for the forward and reverse read, respectively.
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    """
    fastqs_fps = self.inputs[self.sam_pool][(tech, self.sam_pool)]
    status_update(self, tech, fastqs_fps)

    fastq_fp = fastqs_fps[num - 1]
    line = get_fastq_header(
        self.config.fastq[self.sam_pool][(tech, self.sam_pool)][num - 1])
    line_split = line.split()
    if not line_split[0].endswith('/%s' % num):
        if len(line_split) > 1:
            cmd = edit_fastq_cmd(fastq_fp, num)
            self.outputs['cmds'].setdefault((tech,), []).append(cmd)
            self.soft.add_status(tech, self.sam_pool, 1)
        else:
            self.soft.add_status(tech, self.sam_pool, 0)
    io_update(self, i_f=fastqs_fps, o_f=fastqs_fps, key=tech)


def edit(self) -> None:
    """Create command lines for editing the fasta if need be so that
    the fastq header all contaim the trailing "/1" and "/2" that allow
    identifying paired reads.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .config
            Configurations
    """
    tech = 'illumina'
    for r in [1, 2]:
        if len(self.config.fastq[self.sam_pool][(tech, self.sam_pool)]) < r:
            continue
        get_edit_cmd(self, r, tech)


def count_cmd(
        self,
        idx: int,
        fastx: str,
        out: str
) -> str:
    """Get the command to count the reads in a fasta, fastq or fastq.gz file.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .soft.params
            Parameters
    idx : int
        Unique, incremental numeric identifier.
    fastx : str
        Path to the input file name.
    out : str
        Path to the output file name.

    Returns
    -------
    cmd : str
        Command to count reads and output the counts.
    """
    if fastx.endswith('fastq.gz'):
        cmd = "n%s=`zcat %s | wc -l | " % (idx, fastx)
        cmd += "sed 's/ //g' | awk '{x=$1/4; print x}'`\n"
        # cmd += "cut -d ' ' -f 1 | awk '{x=$1/4; print x}'`"
    elif fastx.endswith('fasta'):
        cmd = "n%s=`wc -l %s | " % (idx, fastx)
        cmd += "sed 's/ //g' | awk '{x=$1/2; print x}'`\n"
        # cmd += "cut -d ' ' -f 1 | awk '{x=$1/2; print x}'`"
    elif fastx.endswith('fastq'):
        cmd = "n%s=`wc -l %s | " % (idx, fastx)
        cmd += "sed 's/ //g' | awk '{x=$1/4; print x}'`\n"
        # cmd += "cut -d ' ' -f 1 | awk '{x=$1/4; print x}'`"
    else:
        raise IOError("Input sequence file invalid %s" % fastx)
    if idx:
        cmd += 'echo "%s,2,${n%s}" >> %s\n' % (self.sam_pool, idx, out)
    else:
        cmd += 'echo "%s,1,${n%s}" > %s\n' % (self.sam_pool, idx, out)
    return cmd


def count(self) -> None:
    """Create command lines for counting the reads.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for MIDAS
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    outs = {}
    for (tech, sam), fastxs in self.inputs[self.sam_pool].items():
        if tech_specificity(self, fastxs, tech, sam):
            continue
        to_dos = status_update(self, tech, fastxs)

        out_dir = '%s/%s/%s' % (self.dir, tech, sam)
        self.outputs['dirs'].append(out_dir)

        out = '%s/read_count.tsv' % out_dir
        outs[(tech, self.sam_pool)] = out

        if self.config.force or to_do(out):
            for idx, fastx in enumerate(fastxs):
                if to_dos:
                    cmd = count_cmd(self, idx, fastx, out)
                    self.outputs['cmds'].setdefault((tech,), []).append(cmd)
                else:
                    self.outputs['cmds'].setdefault((tech,), []).append(False)
            io_update(self, i_f=fastxs, i_d=out_dir, o_f=out, key=tech)
            self.soft.add_status(tech, sam, 1)
        else:
            self.soft.add_status(tech, sam, 0)
    self.outputs['outs'] = outs


def fastqc(self) -> None:
    """FastQC is a program designed to spot potential problems in high
    througput sequencing datasets. It runs a set of analyses on one or more
    raw sequence files in fastq or bam format and produces a report which
    summarises the results.

    References
    ----------
    Andrews, S. (2010). FastQC:  A Quality Control Tool for High Throughput
    Sequence Data [Online].

    Notes
    -----
    GitHub  : https://github.com/s-andrews/FastQC
    Docs    : https://www.bioinformatics.babraham.ac.uk/projects/fastqc/

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for FastQC
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    outs = {}
    for (tech, sam), fastxs in self.inputs[self.sam_pool].items():
        if tech_specificity(self, fastxs, tech, sam):
            continue
        to_dos = status_update(self, tech, fastxs)

        out_dir = '%s/%s/%s' % (self.dir, tech, self.sam_pool)
        self.outputs['dirs'].append(out_dir)

        out = ['%s/%s_fastqc.html' % (
            out_dir, basename(x).rsplit('.fastq', 1)[0]) for x in fastxs]
        outs[(tech, self.sam_pool)] = out

        if self.config.force or sum([to_do(x) for x in out]):
            cmd = 'fastqc %s -o %s' % (' '.join(fastxs), out_dir)
            if to_dos:
                self.outputs['cmds'][(tech,)] = [False]
            else:
                self.outputs['cmds'][(tech,)] = [cmd]
            io_update(self, i_f=fastxs, o_d=out_dir, key=tech)
            self.soft.add_status(tech, sam, 1)
        else:
            self.soft.add_status(tech, sam, 0)

    self.outputs['outs'] = outs


def fastp_cmd(
        self,
        tech: str,
        fastqs: list,
        out_dir: str
) -> tuple:
    """Collect the Fastp command.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .soft.params
            Parameters
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastqs : list
        Path to the input files
    out_dir : str
        Path to the output folder

    Returns
    -------
    cmd : str
        Fastp command
    outs : list
        Path to the output files
    """
    params = tech_params(self, tech)
    outs = []
    cmd = 'fastp'
    for fdx, fastq in enumerate(fastqs):
        if fdx:
            i, o = 'I', 'O'
            out = '%s/%s_2.fastq' % (out_dir, self.sam_pool)
        else:
            i, o = 'i', 'o'
            out = '%s/%s_1.fastq' % (out_dir, self.sam_pool)
        if fastq.endswith('.gz'):
            out += '.gz'
        outs.append(out)
        cmd += '  -%s %s -%s %s' % (i, fastq, o, out)
    cmd += ' --thread %s' % params['cpus']
    cmd += ' --report_title="%s"' % self.sam_pool
    cmd += ' --json=%s/%s.json' % (out_dir, self.sam_pool)
    cmd += ' --html=%s/%s.html' % (out_dir, self.sam_pool)

    if params['split_by_lines']:
        cmd += ' --split_by_lines'
        cmd += ' --split_prefix_digits %s' % params[
            'split_prefix_digits']
    elif params['split']:
        cmd += ' --split'
        cmd += ' --split_prefix_digits %s' % params[
            'split_prefix_digits']

    if params['overrepresentation_analysis']:
        cmd += ' --overrepresentation_analysis'
        cmd += ' --overrepresentation_sampling %s' % params[
            'overrepresentation_sampling']

    if params['correction']:
        cmd += ' --correction'
        cmd += ' --overlap_len_require %s' % params[
            'overlap_len_require']
        cmd += ' --overlap_diff_limit %s' % params[
            'overlap_diff_limit']

    if params['disable_length_filtering']:
        cmd += ' --disable_length_filtering'
    else:
        cmd += ' --length_required %s' % params['length_required']
        cmd += ' --length_limit %s' % params['length_limit']

    if params['disable_adapter_trimming']:
        cmd += ' --disable_adapter_trimming'
    else:
        if params['detect_adapter_for_pe']:
            cmd += ' --detect_adapter_for_pe'
        if params['adapter_fasta']:
            cmd += ' --adapter_fasta %s' % params['adapter_fasta']
        else:
            for adapter in ['adapter_sequence', 'adapter_sequence_r2']:
                if params[adapter]:
                    cmd += ' --%s %s' % (adapter, params[adapter])

    if params['disable_quality_filtering']:
        cmd += ' --disable_quality_filtering'
    else:
        for param in ['qualified_quality_phred', 'unqualified_percent_limit',
                      'n_base_limit', 'average_qual']:
            cmd += ' --%s %s' % (param, params[param])

    for param in ['compression', 'trim_front1', 'trim_tail1', 'max_len1',
                  'trim_front2', 'trim_tail2', 'max_len2',
                  'dup_calc_accuracy', 'cut_window_size', 'cut_mean_quality']:
        if params[param]:
            cmd += ' --%s %s' % (param, params[param])

    for boolean in ['dont_overwrite', 'phred64', 'verbose', 'dedup',
                    'dont_eval_duplication', 'trim_poly_g', 'trim_poly_x',
                    'cut_front', 'cut_tail', 'cut_right']:
        if params[boolean]:
            if boolean == 'trim_poly_g' and not params['disable_trim_poly_g']:
                cmd += ' --%s' % boolean
                cmd += ' --poly_g_min_len %s' % params['poly_g_min_len']
            else:
                cmd += ' --%s' % boolean
                if boolean == 'trim_poly_x':
                    cmd += ' --poly_x_min_len %s' % params['poly_x_min_len']
                elif boolean.startswith('cut_'):
                    window_size = '%s_window_size' % boolean
                    mean_quality = '%s_mean_quality' % boolean
                    cmd += ' --%s %s' % (window_size, params[window_size])
                    cmd += ' --%s %s' % (mean_quality, params[mean_quality])
            if boolean == 'cut_front':
                cmd += ' --cut_front_mean_quality %s' % (
                    params['cut_front_mean_quality'])
                cmd += ' --cut_front_window_size %s' % (
                    params['cut_front_window_size'])
            if boolean == 'cut_tail':
                cmd += ' --cut_tail_mean_quality %s' % (
                    params['cut_tail_mean_quality'])
                cmd += ' --cut_tail_window_size %s' % (
                    params['cut_tail_window_size'])
            if boolean == 'cut_right':
                cmd += ' --cut_right_mean_quality %s' % (
                    params['cut_right_mean_quality'])
                cmd += ' --cut_right_window_size %s' % (
                    params['cut_right_window_size'])

    return cmd, outs


def fastp(self) -> None:
    """A tool designed to provide fast all-in-one preprocessing for FastQ
    files. This tool is developed in C++ with multithreading supported to
    afford high performance.

    References
    ----------
    Chen, Shifu, et al. "fastp: an ultra-fast all-in-one FASTQ preprocessor."
    Bioinformatics 34.17 (2018): i884-i890.

    Notes
    -----
    GitHub  : https://github.com/OpenGene/fastp
    Paper   : https://doi.org/10.1093/bioinformatics/bty560

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for Fastp
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    for (tech, sam), fastqs in self.inputs[self.sam_pool].items():
        if tech_specificity(self, fastqs, tech, sam):
            continue
        to_dos = status_update(self, tech, fastqs)

        out_dir = '%s/%s/%s' % (self.dir, tech, self.sam_pool)
        self.outputs['dirs'].append(out_dir)

        cmd, outs = fastp_cmd(self, tech, fastqs, out_dir)
        self.outputs['outs'].setdefault((tech, self.sam_pool), []).extend(outs)

        if self.config.force or sum([to_do(x) for x in outs]):
            if to_dos:
                self.outputs['cmds'][(tech,)] = [False]
            else:
                self.outputs['cmds'][(tech,)] = [cmd]
            io_update(self, i_f=fastqs, o_d=out_dir, key=tech)
            self.soft.add_status(tech, sam, 1)
        else:
            self.soft.add_status(tech, sam, 0)


def cutadapt_cmd(
        self,
        tech: str,
        fastqs: list,
        out_dir: str
) -> tuple:
    """Collect the Cutadapt command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastqs : list
        Path to the input files
    out_dir : str
        Path to the output folder

    Returns
    -------
    cmd : str
        Cutadapt command
    outs : list
        Path to the output files
    """
    params = tech_params(self, tech)
    cmd = 'cutadapt'
    cmd += ' -a AGATCGGAAGAGCACACGTCTGAACTCCAGTCA'
    cmd += ' -A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGT'
    for boolean in [
        'trim_n', 'revcomp', 'zero_cap', 'no_indels', 'pair_adapters',
        'discard_trimmed', 'discard_untrimmed', 'match_read_wildcards',
        'no_match_adapter_wildcards'
    ]:
        if params[boolean]:
            cmd += ' --%s' % boolean.replace('_', '-')
    for param in [
        'times', 'overlap', 'error_rate', 'quality_base', 'report',
        'pair-filter', 'action'
    ]:
        cmd += ' --%s %s' % (param.replace('_', '-'), params[param])
    cmd += ' --minimum-length 10'
    cmd += ' --cores %s' % params['cpus']

    r1_o = '%s/%s.R1.fastq.gz' % (out_dir, self.sam_pool)
    cmd += ' --output %s' % r1_o
    outs = [r1_o]
    if len(fastqs) == 2:
        r2_o = '%s/%s.R2.fastq.gz' % (out_dir, self.sam_pool)
        cmd += ' --paired-output %s ' % r2_o
        outs.append(r2_o)
    cmd += ' '.join(fastqs)
    return cmd, outs


def cutadapt(self) -> None:
    """Cutadapt finds and removes adapter sequences, primers, poly-A tails
    and other types of unwanted sequence from your high-throughput sequencing
    reads.

    Cleaning your data in this way is often required: Reads from small-RNA
    sequencing contain the 3’ sequencing adapter because the read is longer
    than the molecule that is sequenced. Amplicon reads start with a primer
    sequence. Poly-A tails are useful for pulling out RNA from your sample,
    but often you don’t want them to be in your reads.

    Cutadapt helps with these trimming tasks by finding the adapter or primer
    sequences in an error-tolerant way. It can also modify and filter
    single-end and paired-end reads in various ways. Adapter sequences can
    contain IUPAC wildcard characters. Cutadapt can also demultiplex your reads.

    References
    ----------
    Martin, Marcel. "Cutadapt removes adapter sequences from high-throughput
    sequencing reads." EMBnet. journal 17.1 (2011): 10-12.

    Notes
    -----
    GitHub  : https://github.com/marcelm/cutadapt
    Docs    : https://cutadapt.readthedocs.io/en/stable/develop.html
    Paper   : https://doi.org/10.14806/ej.17.1.200

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for Cutadapt
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    for (tech, sam), fastqs in self.inputs[self.sam_pool].items():
        if tech_specificity(self, fastqs, tech, sam, ['illumina']):
            continue
        to_dos = status_update(self, tech, fastqs)

        out_dir = '%s/%s/%s' % (self.dir, tech, self.sam_pool)
        self.outputs['dirs'].append(out_dir)

        cmd, outs = cutadapt_cmd(self, tech, fastqs, out_dir)
        self.outputs['outs'].setdefault((tech, self.sam_pool), []).extend(outs)

        if self.config.force or sum([to_do(x) for x in outs]):
            if to_dos:
                self.outputs['cmds'][(tech,)] = [False]
            else:
                self.outputs['cmds'][(tech,)] = [cmd]
            io_update(self, i_f=fastqs, o_f=outs, key=tech)
            self.soft.add_status(tech, sam, 1)
        else:
            self.soft.add_status(tech, sam, 0)


def atropos_outs(
        self,
        fastqs,
        out_dir: str
) -> list:
    """Collect the output file paths for Atropos.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
    fastqs : list
        Path to the input files
    out_dir : str
        Path to the output folder

    Returns
    -------
    outs : list
        Path to the output files
    """
    if len(fastqs) == 1:
        outs = ['%s/%s.fastq' % (out_dir, self.sam_pool)]
    else:
        outs = ['%s/%s_R1.fastq' % (out_dir, self.sam_pool),
                '%s/%s_R2.fastq' % (out_dir, self.sam_pool)]
    outs = ['%s.gz' % outs[idx] if x.endswith('.gz') else x
            for idx, x in enumerate(fastqs)]
    return outs


def atropos_cmd(
        self,
        tech: str,
        inputs: list,
        out_dir: str,
        outputs: list
) -> str:
    """Collect the command line for Atropos filtering and trimming.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .soft.params
            Parameters
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    inputs : list
        Path to the input files
    out_dir : str
        Path to the output folder
    outputs : list
        Path to the output files

    Returns
    -------
    cmd : str
        Atropos commands
    """
    params = tech_params(self, tech)
    cmd = 'atropos trim'
    for a in params['a']:
        cmd += ' -a %s' % a
    for A in params['A']:
        cmd += ' -A %s' % A
    cmd += ' -q %s' % params['q']
    cmd += ' --minimum-length %s' % params['minimum_length']
    cmd += ' --nextseq-trim %s' % params['nextseq_trim']
    cmd += ' --pair-filter %s' % params['pair_filter']
    cmd += ' --overlap %s' % params['overlap']
    cmd += ' --max-reads %s' % params['max_reads']
    cmd += ' --indel-cost %s' % params['indel_cost']
    cmd += ' --error-rate %s' % params['error_rate']
    cmd += ' --quality-cutoff %s' % params['quality_cutoff']
    cmd += ' --threads %s' % params['cpus']
    cmd += ' --report-file %s/%s.log' % (out_dir, self.sam_pool)
    cmd += ' --report-formats %s' % params['report_formats']
    if len(inputs) == 1:
        cmd += ' -se %s' % inputs[0]
        cmd += ' -o %s' % outputs[0]
    else:
        cmd += ' -pe1 %s' % inputs[0]
        cmd += ' -pe2 %s' % inputs[1]
        cmd += ' -o %s' % outputs[0]
        cmd += ' -p %s' % outputs[1]
    return cmd


def atropos(self) -> None:
    """Atropos is tool for specific, sensitive, and speedy trimming of NGS
    reads. It is a fork of the venerable Cutadapt read trimmer.

    References
    ----------
    Didion, John P., Marcel Martin, and Francis S. Collins. "Atropos:
    specific, sensitive, and speedy trimming of sequencing reads." PeerJ 5 (
    2017): e3720.

    Notes
    -----
    GitHub  : https://github.com/jdidion/atropos
    Docs    : https://atropos.readthedocs.io/en/1.1/
    Paper   : https://doi.org/10.7717/peerj.3720

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for Atropos
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    for (tech, sam), fastqs in self.inputs[self.sam_pool].items():
        if tech_specificity(self, fastqs, tech, sam, ['illumina']):
            continue
        to_dos = status_update(self, tech, fastqs)

        out_dir = '%s/%s/%s' % (self.dir, tech, self.sam_pool)
        self.outputs['dirs'].append(out_dir)
        outs = atropos_outs(self, fastqs, out_dir)

        cmd = atropos_cmd(self, tech, fastqs, out_dir, outs)
        if self.config.force or sum([to_do(x) for x in outs]):
            if to_dos:
                self.outputs['cmds'][(tech,)] = [False]
            else:
                self.outputs['cmds'][(tech,)] = [cmd]
            io_update(self, i_f=fastqs, o_f=outs, key=tech)
            self.soft.add_status(tech, sam, 1)
        else:
            self.soft.add_status(tech, sam, 0)

        self.outputs['outs'].setdefault((tech, self.sam_pool), []).extend(outs)


def hifiadapterfilt_cmd(
        self,
        fastqs: list,
        out_dir: str,
        sam: str
) -> str:
    """Collect the command line for HiFiAdapterFilt.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .soft.params
            Parameters
    fastqs : list
        Path to the input files
    out_dir : str
        Path to the output folder
    sam : str
        Name of the sample (prefix)

    Returns
    -------
    cmd : str
        HiFiAdapterFilt commands
    """
    cmd = 'export PATH=$PATH:%s\n' % self.soft.params['path']
    cmd += 'export PATH=$PATH:%s/DB\n' % self.soft.params['path']
    cmd += 'cd %s\n' % out_dir
    cmd += 'cp %s %s/.\n' % (' '.join(fastqs), out_dir)
    cmd += 'bash %s/pbadapterfilt.sh' % self.soft.params['path']
    cmd += ' -p %s' % sam
    cmd += ' -l %s' % self.soft.params['l']
    cmd += ' -m %s' % self.soft.params['m']
    cmd += ' -t %s' % self.soft.params['cpus']
    cmd += ' -o %s\n' % out_dir
    for fastq in fastqs:
        cmd += 'rm %s/%s\n' % (out_dir, basename(fastq))
    return cmd


def hifiadapterfilt(self) -> None:
    """Convert .bam to .fastq and remove reads with remnant PacBio adapter
    sequences.

    References
    ----------
    Sim, Sheina B., et al. "HiFiAdapterFilt, a memory efficient read
    processing pipeline, prevents occurrence of adapter sequence in PacBio
    HiFi reads and their negative impacts on genome assembly." BMC genomics
    23.1 (2022): 1-7.

    Notes
    -----
    GitHub  : https://github.com/sheinasim/HiFiAdapterFilt
    Paper   : https://doi.org/10.1186/s12864-022-08375-1

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for HiFiAdapterFilt
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    for (tech, sam), fastqs in self.inputs[self.sam_pool].items():
        if tech_specificity(self, fastqs, tech, sam, ['pacbio']):
            continue
        to_dos = status_update(self, tech, fastqs)

        out_dir = '%s/%s/%s' % (self.dir, tech, self.sam_pool)
        self.outputs['dirs'].append(out_dir)

        out = '%s/%s.filt.fastq.gz' % (out_dir, sam)
        self.outputs['outs'].setdefault((tech, self.sam_pool), []).append(out)

        if self.config.force or to_do(out):
            cmd = hifiadapterfilt_cmd(self, fastqs, out_dir, sam)
            key = (tech, sam)
            if to_dos:
                self.outputs['cmds'].setdefault(key, []).append(False)
            else:
                self.outputs['cmds'].setdefault(key, []).append(cmd)
            io_update(self, i_f=fastqs, o_d=out_dir, key=key)
            self.soft.add_status(tech, sam, 1)
        else:
            self.soft.add_status(tech, sam, 0)


def kneaddata_cmd(
        self,
        tech: str,
        inputs: list,
        out_dir: str
) -> tuple:
    """Collect the Kneaddata commands.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .outputs : dict
            All outputs
        .soft.params
            Parameters
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    inputs : list
        Path to the input files
    out_dir : str
        Path to the output folder

    Returns
    -------
    cmd : str
        Kneaddata commands
    outputs : list
        Path to the output files
    """
    params = tech_params(self, tech)
    cmd = 'kneaddata'
    for inp in inputs:
        cmd += ' --input %s' % inp
    for database in params['databases']:
        cmd += ' --reference-db %s' % database
    cmd += ' --output %s' % out_dir
    cmd += ' --processes %s' % params['cpus']
    cmd += ' --max-memory %s%s' % (params['mem'], params['mem_dim'][0])
    if params['trimmomatic']:
        cmd += ' --trimmomatic %s' % params['trimmomatic']
    if params['bowtie2']:
        cmd += ' --bowtie2 %s' % params['bowtie2']
    cmd += ' --bowtie2-options="--very-sensitive-local -p %s"' % params['cpus']
    cmd += ' --remove-intermediate-output'
    cmd += ' --log %s/%s.log\n' % (out_dir, self.sam_pool)
    # write these additional output files manipulations commands
    rad = basename(inputs[0]).replace('.fastq.gz', '')
    outputs = []
    for typ in ['paired', 'unmatched']:
        for orient in ['1', '2']:
            out = '%s/%s_kneaddata_%s_%s.fastq' % (out_dir, rad, typ, orient)
            r = out.replace('.R1', '')
            self.outputs['outs'].setdefault((tech, self.sam_pool), []).append(r)
            if out == r:
                continue
            outputs.append(r)
            cmd += 'mv %s %s\n' % (out, r)
    if params['purge']:
        cmd += 'rm %s/%s*\n' % (out_dir, rad)
    else:
        cmd += 'tar cpfz %s/%s_kneadfiles.tar.gz %s/%s* --remove-files\n' % (
            out_dir, self.sam_pool, out_dir, rad)
    return cmd, outputs


def kneaddata(self) -> None:
    """KneadData is a tool designed to perform quality control on metagenomic
    and metatranscriptomic sequencing data, especially data from microbiome
    experiments. In these experiments, samples are typically taken from a
    host in hopes of learning something about the microbial community on the
    host. However, sequencing data from such experiments will often contain a
    high ratio of host to bacterial reads. This tool aims to perform
    principled in silico separation of bacterial reads from these
    "contaminant" reads, be they from the host, from bacterial 16S sequences,
    or other user-defined sources. Additionally, KneadData can be used for
    other filtering tasks. For example, if one is trying to clean data
    derived from a human sequencing experiment, KneadData can be used to
    separate the human and the non-human reads.

    Notes
    -----
    GitHub  : https://github.com/biobakery/kneaddata
    Docs    : https://huttenhower.sph.harvard.edu/kneaddata

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for Kneaddata
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    for (tech, sam), fastqs in self.inputs[self.sam_pool].items():
        if tech_specificity(self, fastqs, tech, sam, ['illumina']):
            continue
        to_dos = status_update(self, tech, fastqs)

        out_dir = '%s/%s/%s' % (self.dir, tech, self.sam_pool)
        self.outputs['dirs'].append(out_dir)

        cmd, outputs = kneaddata_cmd(self, tech, fastqs, out_dir)

        if self.config.force or sum([to_do(x) for x in outputs]):
            if to_dos:
                self.outputs['cmds'][(tech,)] = [False]
            else:
                self.outputs['cmds'][(tech,)] = [cmd]
            io_update(self, i_f=fastqs, o_d=out_dir, key=tech)
            self.soft.add_status(tech, sam, 1)
        else:
            self.soft.add_status(tech, sam, 0)


def filtering_cmd(
        self,
        tech: str,
        db_path: str,
        inputs: list,
        out1: str,
        out2: str
) -> tuple:
    """Collect the command line for filtering.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    db_path : str
        Path to the reference bowtie2 database
    inputs : list
        Path to the input fastq files
    out1 : str
        Path to the output fastq.gz files
    out2 : str
        Path to the output fastq.gz files

    Returns
    -------
    cmd : str
        filtering commands
    fastqs : list
        Path to the output fastq files
    """
    params = tech_params(self, tech)
    cmd = '\nbowtie2'
    cmd += ' -p %s' % params['cpus']
    cmd += ' -x %s' % db_path
    cmd += ' --very-sensitive'
    if len(inputs) == 1:
        cmd += ' -U %s' % inputs[0]
    else:
        cmd += ' -1 %s' % inputs[0]
        cmd += ' -2 %s' % inputs[1]
    cmd += ' | samtools view -f 12 -F 256'
    cmd += ' | samtools sort -@ %s -n' % params['cpus']
    cmd += ' | samtools view -bS'
    cmd += ' | bedtools bamtofastq -i -'
    cmd += ' -fq %s' % out1.replace('fastq.gz', 'fastq')
    fastqs = [out1.replace('fastq.gz', 'fastq')]
    if len(inputs) > 1:
        cmd += ' -fq2 %s' % out2.replace('fastq.gz', 'fastq')
        fastqs.append(out2.replace('fastq.gz', 'fastq'))
    cmd += '\n'
    return cmd, fastqs


def filtering(self):
    """Perform the filtering of fastq files so that every sequences that
    aligns to any of the passed databases is discarded.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for filtering
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    for (tech, sam), fastqs_ in self.inputs[self.sam_pool].items():
        if tech_specificity(self, fastqs_, tech, sam):
            continue
        to_dos = status_update(self, tech, fastqs_)

        fastqs = fastqs_
        cmds = ''
        out_dirs = []
        databases = self.soft.params['databases']
        for dx, (db, db_path) in enumerate(databases.items()):
            inputs = list(fastqs)
            out_dir = '%s/%s/%s_%s/%s' % (self.dir, tech, dx, db, self.sam_pool)
            out_dirs.append(out_dir)
            self.outputs['dirs'].append(out_dir)
            if len(inputs) == 1:
                out1 = '%s/%s.fastq' % (out_dir, self.sam_pool)
            else:
                out1 = '%s/%s_R1.fastq' % (out_dir, self.sam_pool)
            out2 = '%s/%s_R2.fastq' % (out_dir, self.sam_pool)
            if inputs[0].endswith('.gz'):
                out1 += '.gz'
                out2 += '.gz'
            cmd, fastqs = filtering_cmd(self, tech, db_path, inputs, out1, out2)
            if self.config.force or to_do(out1):
                cmds += cmd
                if dx:
                    cmds += '\nrm %s' % ' '.join(inputs)
        if fastqs[0].endswith('.fastq'):
            for fastq in fastqs:
                cmds += '\ngzip -f %s' % fastq
        fastqs_gz = ['%s.gz' % x for x in fastqs]
        self.outputs['outs'].setdefault(
            (tech, self.sam_pool), []).extend(fastqs_gz)
        if (self.config.force or sum([to_do(x) for x in fastqs_gz])) and cmds:
            if to_dos:
                self.outputs['cmds'][(tech,)] = [False]
            else:
                self.outputs['cmds'][(tech,)] = [cmds]
            io_update(self, i_f=fastqs_, i_d=out_dirs, o_f=fastqs_gz, key=tech)
            self.soft.add_status(tech, sam, 1)
        else:
            self.soft.add_status(tech, sam, 0)


def mapdamage2(self) -> None:
    """.

    Notes
    -----
    GitHub  : https://ginolhac.github.io/mapDamage
    paper   : https://doi.org/10.1093/bioinformatics/btt193

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for MapDamage2
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    pass



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import datetime as dt
from os.path import abspath, basename, isdir

from metagenomix.metagenomix import metagenomix
from metagenomix._io_utils import print_status_table
from metagenomix.core.output import Output


def monitoring(**kwargs):
    """Show the status of the planned outputs."""

    print('\n>>> `metagenomix monitor` started >>>\n')

    # Collect all command and init the script creating instance
    monitor = Monitored(**kwargs)

    print('* setting up the output file name')
    monitor.make_status_dir()
    monitor.get_out()

    print('* collecting and showing the current status of the analyses')
    monitor.monitor_status()
    monitor.write_status()

    monitor.parse_softs()
    monitor.monitor_softs()

    print('\n<<< `metagenomix monitor` completed <<<\n')


class Monitored(object):

    def __init__(self, **kwargs):
        kwargs['localscratch'] = None
        kwargs['userscratch'] = False
        kwargs['purge_pfams'] = None
        kwargs['show_pfams'] = None
        kwargs['scratch'] = False
        kwargs['chunks'] = None
        kwargs['jobs'] = None
        config, databases, workflow, commands = metagenomix(**kwargs)
        self.__dict__.update(kwargs)
        self.config = config
        self.databases = databases
        self.graph = workflow.graph
        self.commands = commands
        self.output_dir = abspath(self.output_dir)
        # self.softs = {'res': {}, 'dir': set(), 'pip': set(), 'usr': set()}
        # USE THE SOFTWARES OF THE PARSED COMMANDS----
        self.time = dt.datetime.now().strftime("%d-%m-%Y_%H-%M")
        self.log_dir = '%s/_monitored' % config.dir
        self.roles = {}
        self.monitored = {}

    def monitor_status(self):
        m = max(len(x) for x in self.commands.softs) + 1
        for sdx, (name, soft) in enumerate(self.commands.softs.items()):
            n = (m - len(name) - len(str(sdx))) + 1
            cur_soft = '%s [%s]' % (sdx, name)
            soft.tables.append(cur_soft)
            print('\t%s %s%s' % (cur_soft, ('.' * n), ('.' * 8)), end=' ')
            print_status_table(soft, True)

    def make_status_dir(self):
        if not isdir(self.log_dir):
            os.makedirs(self.log_dir)

    def get_out(self):
        if self.summary_fp is None:
            base = self.time + '.txt'
        else:
            base = basename(self.summary_fp)
            if '/' in self.summary_fp:
                print('Using "%s" to write in "%s"' % (base, self.log_dir))
        self.summary_fp = self.log_dir + '/' + base

    def write_status(self):
        with open(self.summary_fp, 'w') as o:
            o.write('# Summary of the data that is currently needed as input\n')
            o.write('# or not yet produced as output.\n')
            o.write('# This file format will evolve...\n')
            o.write('# Date of status summary: %s\n' % self.time)
            for sdx, (name, soft) in enumerate(self.commands.softs.items()):
                # hashed = self.get_hash(soft)
                o.write('\n%s\n' % '\t'.join(soft.tables[:2]))
                for table in soft.tables[2:]:
                    if table is None:
                        o.write(' -> All necessary data available\n')
                    else:
                        o.write('%s\n' % table)
        print('Written: %s' % self.summary_fp)

    def parse_softs(self):
        """An Output class instance is created for each software to manage,
        and placed as value to the dict with the software name of key."""
        for name, soft in self.commands.softs.items():
            if isdir(self.output_dir + '/' + name):
                output = Output(self.output_dir, name)
                output.get_afters()
                output.get_outputs()
                self.monitored[name] = output.outputs

    def monitor_softs(self):
        for name, outputs in self.monitored.items():
            print()
            print()
            print()
            print()
            print()
            print('#' * 40)
            print('software:', name)
            print('#' * 40)
            for (after, hash_value), data in outputs.items():
                print()
                print()
                print("after, hash_value", after, hash_value)
                for k, d in data.items():
                    print()
                    print(k)
                    print(d)
            print(sdkjvdsb)
            # print(pd.DataFrame(soft.jobs))




# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import sys
from os.path import dirname
from metagenomix._inputs import (sample_inputs, group_inputs,
                                 genome_key, genome_out_dir)
from metagenomix._io_utils import io_update, to_do, status_update


def plasforest_cmd(
        self,
        in_fp: str,
        out_fp: str,
        out_dir: str
) -> str:
    """Collect plasforest command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    in_fp : str
        Path to the input file
    out_fp : str
        Path to the output file
    out_dir : str
        Path to the output folder

    Returns
    -------
    cmd : str
        plasforest command
    """
    binary = self.soft.params['binary']
    cmd = 'cd %s\n' % out_dir
    cmd += 'cp %s/plasforest.sav %s/.\n' % (dirname(binary), out_dir)
    cmd += 'cp %s/*.fasta* %s/.\n' % (dirname(binary), out_dir)
    cmd += 'python3 %s' % binary
    cmd += ' -i %s' % in_fp
    cmd += ' -o %s' % out_fp
    if self.soft.params['size_of_batch']:
        cmd += ' --size_of_batch %s' % self.soft.params['size_of_batch']
    cmd += ' --threads %s' % self.soft.params['cpus']
    for boolean in ['b', 'f', 'r']:
        if self.soft.params[boolean]:
            cmd += ' -%s' % boolean
    cmd += '\nrm %s/plasforest.sav\n' % out_dir
    cmd += 'rm %s/*.fasta*\n' % out_dir
    return cmd


def get_plasforest(
        self,
        tech: str,
        fastas: dict,
        sam_group: str
) -> None:
    """

    Parameters
    ----------
    self : Commands class instance
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastas : dict
        Paths to the input fasta files per genome/MAG
    sam_group : str
        Sample or co-assembly name
    """
    for genome, fasta in fastas.items():

        out_dir = genome_out_dir(self, tech, fasta[0], sam_group, genome)
        self.outputs['dirs'].append(out_dir)

        out_fp = '%s/plasmids.csv' % out_dir
        self.outputs['outs'].setdefault((tech, sam_group), []).append(out_fp)
        to_dos = status_update(self, tech, [fasta[0]], group=sam_group,
                               genome=genome)

        if self.config.force or to_do(out_fp):
            key = genome_key(tech, sam_group, genome)
            cmd = plasforest_cmd(self, fasta[0], out_fp, out_dir)
            if to_dos:
                self.outputs['cmds'].setdefault(key, []).append(False)
            else:
                self.outputs['cmds'].setdefault(key, []).append(cmd)
            io_update(self, i_f=fasta[0], i_d=out_dir, o_f=out_fp, key=key)
            self.soft.add_status(
                tech, self.sam_pool, 1, group=sam_group, genome=genome)
        else:
            self.soft.add_status(
                tech, self.sam_pool, 0, group=sam_group, genome=genome)


def plasmidfinder_cmd(
        self,
        fasta: str,
        out_dir: str,
        key: tuple
) -> str:
    """Collect PlasmidFinder command.

    Parameters
    ----------
    self : Commands class instance
        .pool : str
            Pool name.
        .soft.params
            PlasmidFinder parameters
    fasta : str
        Path to the input file
    out_dir : str
        Path to the output folder for the current sample/MAG
    key : str
        Technology and/or co-assembly pool group name

    Returns
    -------
    cmd : str
        PlasmidFinder command
    """
    tmp_dir = '$TMPDIR/plasmidfinder_%s' % '_'.join(key)
    cmd = 'mkdir -p %s\n' % tmp_dir
    cmd += '%s' % self.soft.params['binary']
    if len(fasta) == 2:
        cmd += ' --infile %s' % ' '.join(fasta)
    else:
        cmd += ' --infile %s' % fasta[0]
    cmd += ' --outputPath %s' % out_dir
    cmd += ' --tmp_dir %s' % tmp_dir
    cmd += ' --methodPath %s' % self.soft.params['methodPath']
    cmd += ' --databasePath %s' % self.soft.params['databasePath']
    if 'databases' in self.soft.params:
        cmd += ' --databases %s' % self.soft.params['databases']
    cmd += ' --mincov %s' % self.soft.params['mincov']
    cmd += ' --threshold %s' % self.soft.params['threshold']
    if self.soft.params['extented_output']:
        cmd += ' --extented_output'
    cmd += '\nrm -rf %s\n' % tmp_dir
    return cmd


def get_plasmidfinder(
        self,
        tech: str,
        fastas: dict,
        sam_group: str
) -> None:
    """

    Parameters
    ----------
    self : Commands class instance
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastas : dict
        Paths to the input fasta files per genome/MAG
    sam_group : str
        Sample or co-assembly name
    """
    for genome, fasta in fastas.items():

        out_dir = genome_out_dir(self, tech, fasta[0], sam_group, genome)
        self.outputs['dirs'].append(out_dir)
        self.outputs['outs'].setdefault((tech, sam_group), []).append(out_dir)
        to_dos = status_update(
            self, tech, [fasta[0]], group=sam_group, genome=genome)

        json_fp = '%s/data.json' % out_dir
        if self.config.force or to_do(json_fp):
            key = genome_key(tech, sam_group, genome)
            cmd = plasmidfinder_cmd(self, fasta, out_dir, key)
            if to_dos:
                self.outputs['cmds'].setdefault(key, []).append(False)
            else:
                self.outputs['cmds'].setdefault(key, []).append(cmd)
            io_update(self, i_f=fasta, i_d=out_dir, o_d=out_dir, key=key)
            self.soft.add_status(
                tech, self.sam_pool, 1, group=sam_group, genome=genome)
        else:
            self.soft.add_status(
                tech, self.sam_pool, 0, group=sam_group, genome=genome)


def dispatch(self) -> None:
    """Classify assembly contigs or genomes/MAGs as plasmids or not.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder
        .prev : str
            Previous software in the pipeline
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .config
            Configurations
    """
    __plasmid_tool__ = getattr(sys.modules[__name__], 'get_%s' % self.soft.name)

    if self.sam_pool in self.pools:
        for (tech, group), inputs in self.inputs[self.sam_pool].items():
            fastas = group_inputs(self, inputs)
            __plasmid_tool__(self, tech, fastas, group)

    elif set(self.inputs) == {''}:
        for (tech, mags), inputs in self.inputs[''].items():
            fastas = group_inputs(self, inputs)
            __plasmid_tool__(self, tech, fastas, mags)

    else:
        if self.soft.name == 'plasmidfinder':
            tech_fastas = sample_inputs(self, raw=True)
        else:
            tech_fastas = sample_inputs(self)
        for tech, fastas in tech_fastas.items():
            __plasmid_tool__(self, tech, fastas, self.sam_pool)


def plasmidfinder(self) -> None:
    """PlasmidFinder is a tool for the identification and typing of Plasmid
    Replicons in Whole-Genome Sequencing (WGS).

    References
    ----------
    Carattoli, Alessandra, and Henrik Hasman. "PlasmidFinder and in silico
    pMLST: identification and typing of plasmid replicons in whole-genome
    sequencing (WGS)." Horizontal gene transfer. Humana, New York, NY,
    2020. 285-294.

    Notes
    -----
    BitBucket   : https://bitbucket.org/genomicepidemiology/plasmidfinder
    Paper       : https://doi.org/10.1007/978-1-4939-9877-7_20

    Parameters
    ----------
    self : Commands class instance
        .soft.name : str
            Name of the current software in the pipeline
        .dir : str
            Path to pipeline output folder for PlasmidFinder
        .soft.prev : str
            Previous software in the pipeline
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .config
            Configurations
        .sam_pool
            Sample of co-assembly group name
    """
    dispatch(self)


def plasforest(self) -> None:
    """A random forest classifier of contigs to identify contigs of plasmid
    origin in contig and scaffold genomes.

    References
    ----------
    Pradier, Léa, et al. "PlasForest: a homology-based random forest
    classifier for plasmid detection in genomic datasets." BMC bioinformatics
    22.1 (2021): 1-17.

    Notes
    -----
    GitHub  : https://github.com/leaemiliepradier/PlasForest
    Paper   : https://doi.org/10.1186/s12859-021-04270-w

    Parameters
    ----------
    self : Commands class instance
        .soft.name : str
            Name of the current software in the pipeline
        .dir : str
            Path to pipeline output folder for PlasForest
        .soft.prev : str
            Previous software in the pipeline
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .config
            Configurations
        .sam_pool
            Sample of co-assembly group name
    """
    dispatch(self)


def mobsuite(self) -> None:
    """MOB-suite: Software softwares for clustering, reconstruction and typing of
    plasmids from draft assemblies.

    Plasmids are mobile genetic elements (MGEs), which allow for rapid
    evolution and adaption of bacteria to new niches through horizontal
    transmission of novel traits to different genetic backgrounds. The
    MOB-suite is designed to be a modular set of softwares for the typing and
    reconstruction of plasmid sequences from WGS assemblies.

    The MOB-suite depends on a series of databases which are too large to be
    hosted in git-hub. They can be downloaded or updated by running mob_init
    or if running any of the softwares for the first time, the databases will
    download and initialize automatically if you do not specify an alternate
    database location. However, they are quite large so the first run will
    take a long time depending on your connection and speed of your computer.

    References
    ----------
    Robertson, James, and John HE Nash. "MOB-suite: software softwares for
    clustering, reconstruction and typing of plasmids from draft assemblies."
    Microbial genomics 4.8 (2018).

    Notes
    -----
    GitHub  : https://github.com/phac-nml/mob-suite
    Paper   : https://doi.org/10.1099/mgen.0.000206

    Parameters
    ----------
    self
    """
    print()


# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from os.path import splitext
from metagenomix.core.parameters import tech_params
from metagenomix._io_utils import (io_update, to_do, tech_specificity,
                                   not_paired, status_update)

# Keep line because read mapping alignments will be python scripts
# scripts = pkg_resources.resource_filename('metagenomix', 'resources/scripts')


def flash_cmd(
        self,
        tech: str,
        fastqs: list,
        out: str
) -> str:
    """Get the FLASh merging command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters for FLASh
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastqs : list
        Paths to the input files
    out : str
        Path to the output folder

    Returns
    -------
    cmd : str
        FLASh command
    """
    params = tech_params(self, tech)
    cmd = 'flash %s' % ' '.join(fastqs)
    cmd += ' --min-overlap %s' % params['min_overlap']
    cmd += ' --max-overlap %s' % params['max_overlap']
    cmd += ' --max-mismatch-density %s' % params['mismatch']
    cmd += ' --output-directory %s' % out
    cmd += ' --output-prefix %s' % self.sam_pool
    cmd += ' --threads %s' % params['cpus']
    cmd += ' --compress'
    return cmd


def flash(self) -> None:
    """FLASH (Fast Length Adjustment of SHort reads) is a very fast and
    accurate software tool to merge paired-end reads from next-generation
    sequencing experiments. FLASH is designed to merge pairs of reads when
    the original DNA fragments are shorter than twice the length of reads.
    The resulting longer reads can significantly improve genome assemblies.
    They can also improve transcriptome assembly when FLASH is used to merge
    RNA-seq data.

    References
    ----------
    Magoč, Tanja, and Steven L. Salzberg. "FLASH: fast length adjustment of
    short reads to improve genome assemblies." Bioinformatics 27.21 (2011):
    2957-2963.

    Notes
    -----
    SourceForge : https://sourceforge.net/projects/flashpage/files/
    GitHub      : https://github.com/genome-vendor/FLASH
    Docs        : https://ccb.jhu.edu/software/FLASH/
    Paper       : https://doi.org/10.1093/bioinformatics/btr507

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for FLASh
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters for FLASh
        .config
            Configurations
    """
    for (tech, sam), fastqs in self.inputs[self.sam_pool].items():
        if tech_specificity(self, fastqs, tech, sam, ['illumina']):
            continue
        if not_paired(self, tech, sam, fastqs):
            continue
        to_dos = status_update(self, tech, fastqs)

        out = '%s/%s/%s' % (self.dir, tech, self.sam_pool)
        self.outputs['dirs'].append(out)

        rad = out + '/' + self.sam_pool
        ext = '%s.extendedFrags.fastq.gz' % rad
        nc1 = '%s.notCombined_1.fastq.gz' % rad
        nc2 = '%s.notCombined_2.fastq.gz' % rad
        outs = [ext, nc1, nc2]
        self.outputs['outs'].setdefault((tech, self.sam_pool), []).extend(outs)

        if self.config.force or sum([to_do(x) for x in outs]):
            cmd = flash_cmd(self, tech, fastqs, out)
            if to_dos:
                self.outputs['cmds'][(tech,)] = [False]
            else:
                self.outputs['cmds'][(tech,)] = [cmd]
            io_update(self, i_f=fastqs, o_d=out, key=tech)
            self.soft.add_status(tech, sam, 1)
        else:
            self.soft.add_status(tech, sam, 0)


def ngmerge_cmd(
        self,
        tech: str,
        fastqs: list,
        ext: str,
        log: str,
        fail: str) -> str:
    """Get the NGmerge merging command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters for NGmerge
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastqs : list
        Paths to the input files
    ext : str
        Path to the merged reads output file
    log : str
        Path to the reads merging log file
    fail : str
        Path to the unmerged reads output file

    Returns
    -------
    cmd : str
        NGmerge command
    """
    params = tech_params(self, tech)
    cmd = 'NGmerge'
    cmd += ' -1 %s' % fastqs[0]
    cmd += ' -2 %s' % fastqs[1]
    cmd += ' -o %s' % ext.replace('.gz', '')
    if params['a']:
        cmd += ' -c %s.log' % log
    else:
        cmd += ' -j %s.log' % log
        cmd += ' -f %s' % fail

    for boolean in ['a', 'd', 's', 'i', 'g']:
        if params[boolean]:
            cmd += ' -%s' % boolean
    for param in ['p', 'm', 'e', 'q', 'u']:
        cmd += ' -%s %s' % (param, params[param])
    cmd += ' -n %s' % params['cpus']
    cmd += ' -z'
    return cmd


def ngmerge(self) -> None:
    """NGmerge operates on paired-end high-throughput sequence reads in two
    distinct modes.
    - In the default stitch mode, NGmerge combines paired-end reads that
      overlap into a single read that spans the full length of the original
      DNA fragment (Fig. 1A). The ends of the merged read are defined by the
      5' ends of the original reads. Reads that fail the stitching process
      (due to a lack of sufficient overlap, or excessive sequencing errors)
      are placed into secondary output files, if the user requires them.
    - The alternative adapter-removal mode returns the original reads as
      pairs, removing the 3' overhangs of those reads whose valid stitched
      alignment has this characteristic (Fig. 1B). Reads whose alignments do
      not have such overhangs (or do not align at all) will also be printed
      to the output files, unmodified.

    References
    ----------
    Gaspar, John M. "NGmerge: merging paired-end reads via novel
    empirically-derived models of sequencing errors." BMC bioinformatics 19.1
    (2018): 1-9.

    Notes
    -----
    GitHub  : https://github.com/harvardinformatics/NGmerge
    Paper   : https://doi.org/10.1186/s12859-018-2579-2

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for NGmerge
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters for NGmerge
        .config
            Configurations
    """
    for (tech, sam), fastqs in self.inputs[self.sam_pool].items():
        if tech_specificity(self, fastqs, tech, sam, ['illumina']):
            continue
        if not_paired(self, tech, sam, fastqs):
            continue
        to_dos = status_update(self, tech, fastqs)

        out = '%s/%s/%s' % (self.dir, tech, self.sam_pool)
        self.outputs['dirs'].append(out)

        rad = out + '/' + self.sam_pool
        if self.soft.params['a']:
            ext = '%s.trim' % rad
            log = '%s_trim' % rad
        else:
            ext = '%s.extendedFrags.fastq.gz' % rad
            log = '%s_merging' % rad
        fail = '%s.notCombined' % rad
        nc1 = '%s.notCombined_1.fastq.gz' % rad
        nc2 = '%s.notCombined_2.fastq.gz' % rad
        outs = [ext, nc1, nc2]
        self.outputs['outs'].setdefault((tech, self.sam_pool), []).extend(outs)

        if self.config.force or sum([to_do(x) for x in outs]):
            cmd = ngmerge_cmd(self, tech, fastqs, ext, log, fail)
            if to_dos:
                self.outputs['cmds'][(tech,)] = [False]
            else:
                self.outputs['cmds'][(tech,)] = [cmd]
            io_update(self, i_f=fastqs, o_d=out, key=tech)
            self.soft.add_status(tech, sam, 1)
        else:
            self.soft.add_status(tech, sam, 0)


def pear_cmd(
        self,
        tech: str,
        fastqs: list,
        rad: str,
        outs: list,
        outs_: list,
        na: str) -> str:
    """Get the PEAR merging command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters for PEAR
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastqs : list
        Paths to the input files
    rad : str
        Radical of the path to the output files
    outs : list
        Paths to the reads merging output files (after renaming)
    outs_ : list
        Paths to the reads merging output files (before renaming)
    na : str
        Path to the unmerged reads output file

    Returns
    -------
    cmd : str
        PEAR command
    """
    params = tech_params(self, tech)
    cmd = 'pear'
    cmd += ' --forward-fastq %s' % fastqs[0]
    cmd += ' --reverse-fastq %s' % fastqs[1]
    cmd += ' --output %s' % rad
    for param in [
        'min_assembly_length', 'max_assembly_length', 'quality_threshold',
        'min_trim_length', 'score_method', 'min_overlap', 'test_method',
        'phred_base', 'cap', 'p_value', 'max_uncalled_base'
    ]:
        cmd += ' --%s %s' % (param.replace('_', '-'), params[param])
    for boolean in ['empirical_freqs', 'keep_original', 'stitch', 'nbase']:
        if params[boolean]:
            cmd += ' --%s' % boolean.replace('_', '-')
    cmd += ' --threads %s\n' % params['cpus']
    for (src, snk) in zip(outs_, outs):
        cmd += 'gzip %s\nmv %s %s\n' % (src.replace('.gz', ''), src, snk)
    cmd += 'gzip %s\n' % na.replace('.gz', '')
    return cmd


def pear(self) -> None:
    """PEAR assembles Illumina paired-end reads if the DNA fragment sizes are
    smaller than twice the length of reads. PEAR can assemble 95% of reads
    with 35-bp mean overlap with a false-positive rate of 0.004. PEAR also
    works with multiplexed data sets where the true underlying DNA fragment
    size varies. PEAR has an extremely low false-positive rate of 0.0003 on
    data sets where no overlap exists between the two reads (i.e. when DNA
    fragment sizes are larger than twice the read length).

    References
    ----------
    Zhang, Jiajie, et al. "PEAR: a fast and accurate Illumina Paired-End
    reAd mergeR." Bioinformatics 30.5 (2014): 614-620.

    Notes
    -----
    GitHub  : https://github.com/tseemann/PEAR
    Docs    : http://www.exelixis-lab.org/web/software/pear
    Paper   : https://doi.org/10.1093/bioinformatics/btt593

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for PEAR
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters for PEAR
        .config
            Configurations
    """
    for (tech, sam), fastqs in self.inputs[self.sam_pool].items():
        if tech_specificity(self, fastqs, tech, sam, ['illumina']):
            continue
        if not_paired(self, tech, sam, fastqs):
            continue
        to_dos = status_update(self, tech, fastqs)

        out = '%s/%s/%s' % (self.dir, tech, self.sam_pool)
        self.outputs['dirs'].append(out)

        rad = out + '/' + self.sam_pool
        ext_ = '%s.assembled.fastq.gz' % rad
        nc1_ = '%s.unassembled.forward.fastq.gz' % rad
        nc2_ = '%s.unassembled.reverse.fastq.gz' % rad
        outs_ = [ext_, nc1_, nc2_]
        ext = '%s.extendedFrags.fastq.gz' % rad
        nc1 = '%s.notCombined_1.fastq.gz' % rad
        nc2 = '%s.notCombined_2.fastq.gz' % rad
        outs = [ext, nc1, nc2]
        na = '%s.discarded.fastq.gz' % rad
        self.outputs['outs'].setdefault((tech, self.sam_pool), []).extend(outs)

        if self.config.force or sum([to_do(x) for x in outs]):
            cmd = pear_cmd(self, tech, fastqs, rad, outs, outs_, na)
            if to_dos:
                self.outputs['cmds'][(tech,)] = [False]
            else:
                self.outputs['cmds'][(tech,)] = [cmd]
            io_update(self, i_f=fastqs, o_d=out, key=tech)
            self.soft.add_status(tech, sam, 1)
        else:
            self.soft.add_status(tech, sam, 0)


def bbmerge_cmd(
        self,
        tech: str,
        fastqs: list,
        os: list
) -> str:
    """Get the bbmerge.sh merging command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters for bbmerge.sh
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastqs : list
        Paths to the input files
    os : list
        Paths to the output files

    Returns
    -------
    cmd : str
        bbmerge.sh command
    """
    params = tech_params(self, tech)
    cmd = 'bbmerge.sh'
    cmd += ' in1=%s in2=%s' % tuple(fastqs)
    cmd += ' out=%s outu1=%s outu2=%s outinsert=%s outc=%s ihist=%s' % tuple(os)
    if params['strictness']:
        cmd += ' %s=t' % params['strictness']
    else:
        for param in [
            'reads', 'ziplevel', 'trimq','minlength', 'maxlength',
            'minavgquality', 'maxexpectederrors', 'forcetrimleft',
            'forcetrimright', 'forcetrimright2', 'forcetrimmod',
            'mininsert', 'mininsert0', 'minoverlap', 'minoverlap0', 'minq',
            'maxq', 'efilter', 'pfilter', 'kfilter', 'maxratio',
            'ratiomargin', 'ratiooffset', 'maxmismatches',
            'ratiominoverlapreduction', 'minsecondratio', 'margin',
            'mismatches', 'k', 'extend', 'extend2', 'iterations',
            'mindepthseed', 'mindepthextend', 'branchmult1', 'branchmult2',
            'branchlower', 'prealloc', 'prefilter', 'filtermem', 'minprob',
            'minapproxoverlap'
        ]:
            cmd += ' %s=%s' % (param, params[param])
        for boolean in [
            'interleaved', 'nzo', 'showhiststats', 'ordered', 'mix',
            'qtrim', 'qtrim2', 'tbo', 'ooi', 'trimpolya', 'usejni', 'merge',
            'ecco', 'trimnonoverlapping', 'useoverlap', 'entropy', 'ouq',
            'owq', 'usequality', 'iupacton', 'ratiomode', 'forcemerge',
            'flatmode', 'requireratiomatch', 'trimonfailure', 'rem', 'rsem',
            'ecctadpole', 'reassemble', 'removedeadends', 'removebubbles',
            'ibb', 'eccbloom', 'testmerge', 'eoom', 'da',
        ]:
            if params[boolean]:
                cmd += ' %s=t' % boolean
            else:
                cmd += ' %s=f' % boolean
    return cmd


def bbmerge(self) -> None:
    """BBMerge is designed to merge two overlapping paired reads into a
    single read. For example, a 2x150bp read pair with an insert size of
    270bp would result in a single 270bp read. This is useful in amplicon
    studies, as clustering and consensus are far easier with single reads
    than paired reads, and also in assembly, where longer reads allow the
    use of longer kmers (for kmer-based assemblers) or fewer comparisons
    (for overlap-based assemblers). And in either case, the quality of the
    overlapping bases is improved. BBMerge is also capable of error-correcting
    the overlapping portion of reads without merging them, as well as merging
    nonoverlapping reads, if enough coverage is available. BBMerge is the
    fastest and by far the most accurate overlap-based read merger currently
    in existence.

    References
    ----------
    Bushnell, Brian, Jonathan Rood, and Esther Singer. "BBMerge–accurate
    paired shotgun read merging via overlap." PloS one 12.10 (2017): e0185056.

    Notes
    -----
    SourceForge : https://sourceforge.net/projects/bbmap/
    Docs        : https://jgi.doe.gov/data-and-tools/software-tools/bbtools/bb-tools-user-guide/bbmerge-guide/
    Paper       : https://doi.org/10.1371/journal.pone.0185056

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for BBmerge
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters for BBmerge
        .config
            Configurations
    """
    # iterate over the inputs
    for (tech, sam), fastqs in self.inputs[self.sam_pool].items():
        if tech_specificity(self, fastqs, tech, sam, ['illumina']):
            continue
        if not_paired(self, tech, sam, fastqs):
            continue
        to_dos = status_update(self, tech, fastqs)

        # make the output directory
        out = '%s/%s/%s' % (self.dir, tech, self.sam_pool)
        self.outputs['dirs'].append(out)

        # get the expected names of some of the ouptuts:
        # - those you want to collect in 'outs' because they will be future inpt
        # - at least one that will help knowing whether the software already run
        rad = out + '/' + self.sam_pool
        ext = '%s.extendedFrags.fastq.gz' % rad
        nc1 = '%s.notCombined_1.fastq.gz' % rad
        nc2 = '%s.notCombined_2.fastq.gz' % rad
        ins = '%s.inserts.txt' % rad
        kmer = '%s.kmer_cardinality.txt' % rad
        ihist = '%s.insert_length.hist' % rad
        outs = [ext, nc1, nc2]
        outs_cmd = outs + [ins, kmer, ihist]
        self.outputs['outs'].setdefault((tech, self.sam_pool), []).extend(outs)

        # check if the tool already run (or if --force) to allow getting command
        if self.config.force or sum([to_do(x) for x in outs]):
            # collect the command line
            cmd = bbmerge_cmd(self, tech, fastqs, outs_cmd)
            # add is to the 'cmds'
            if to_dos:
                self.outputs['cmds'][(tech,)] = [False]
            else:
                self.outputs['cmds'][(tech,)] = [cmd]
            io_update(self, i_f=fastqs, o_d=out, key=tech)
            self.soft.add_status(tech, sam, 1)
        else:
            self.soft.add_status(tech, sam, 0)


def grep_sam_tail_cmd(
        cmd: str,
        reads: str,
        sam: str
) -> str:
    """Grep the end of the alignment and compare is to end of input to make
    sure that the alignment was not aborted and hence that the output file is
    not incomplete.

    Notes
    -----
    It is possible that the last sequences of the input file do not align,
    but it is safer to recompute the alignment than to assume that and
    potentially leave many sequences out.

    Parameters
    ----------
    cmd : str
        Alignment commands
    reads : str
        last reads of the input sequences file
    sam : str
        Path to the output .sam alignment file

    Returns
    -------
    grep_cmd : str
        Command to grep the end of the alignment and compare is to end of input
    """
    grep_cmd = '\nlast_reads=`%s`\n' % reads
    grep_cmd += 'GREPOK=0\n'
    grep_cmd += 'for i in $listVar\n'
    grep_cmd += 'do\n'
    grep_cmd += '    if grep -q "$i" %s\n' % sam
    grep_cmd += '        then\n'
    grep_cmd += '            GREPOK=1\n'
    grep_cmd += '            break\n'
    grep_cmd += '    fi\n'
    grep_cmd += 'done\n'
    grep_cmd += "if [ $GREPOK = 0 ]\n"
    grep_cmd += 'then\n'
    grep_cmd += '%s\n' % cmd
    grep_cmd += 'fi\n'
    return grep_cmd


def condition_ali_command(
        fastxs: list,
        cmd: str,
        sam: str
) -> str:
    """Add the conditional statements checking that the alignment
    already available was not aborted, in which case the alignment
    will be re-run.

    Parameters
    ----------
    fastxs : list
        Path to the input fasta/fastq/fastq.gz files
    cmd : str
        Alignment command
    sam : str
        Path to the alignment file

    Returns
    -------
    cmd : str
        Alignment command potentially decorated with unix check to rerun
    """
    if fastxs[-1].endswith('fastq.gz'):
        reads = 'zcat %s | tail -n 80 | grep "^@"' % fastxs[-1]
    elif fastxs[-1].endswith('fastq'):
        reads = 'tail -n 80 %s | grep "^@"' % fastxs[-1]
    elif fastxs[-1].endswith('fasta'):
        reads = 'tail -n 40 %s | grep "^>"' % fastxs[-1]
    else:
        return cmd

    cmd = grep_sam_tail_cmd(cmd, reads, sam)
    return cmd


# def get_burst_cmd(self, tech: str, fastx: str, db_path: str, out: str) -> str:
#     """Get the burst alignment command.
#
#     Parameters
#     ----------
#     tech : str
#         Technology: 'illumina', 'pacbio', or 'nanopore'
#     fastx : str
#         Path to the input fasta file
#     db_path : str
#         Path to the burst database
#     out : str
#         Path to the output folder
#
#     Returns
#     -------
#     cmd : str
#         Alignment command
#     """
#     params = tech_params(self, tech)
#     cmd = 'burst'
#     cmd += ' -q %s' % fastx
#     cmd += ' -r %s.edx' % db_path
#     cmd += ' -a %s.acx' % db_path
#     cmd += ' -sa'
#     cmd += ' -fr'
#     cmd += ' -i 0.98'
#     cmd += ' -t %s' % params['cpus']
#     cmd += ' -o %s' % out
#     return cmd


def get_bowtie2_cmd(
        self,
        tech: str,
        fastx: list,
        db_path: str,
        db_out: str
) -> tuple:
    """Get the bowtie2 alignment command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters for humann
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastx : list
        Path to the input fasta/fastq/fastq.gz files
    db_path : str
        Path to the bowtie2 database
    db_out: str
        Path to the output folder

    Returns
    -------
    cmd : str
        Alignment command
    sam : str
        Alignment .sam file
    """
    params = tech_params(self, tech)
    cmd = 'bowtie2'
    cmd += ' -p %s' % params['cpus']
    cmd += ' -x %s' % db_path
    sam = '%s/alignment.bowtie2' % db_out
    if len(fastx) == 2 and params['pairing'] == 'paired':
        cmd += ' -1 %s' % fastx[0]
        cmd += ' -2 %s' % fastx[1]
        if not params['discordant']:
            cmd += ' --no-discordant'
            sam += '.nodiscordant'
    else:
        cmd += ' -U %s' % ','.join(fastx)
    sam += '.sam'
    cmd += ' -S %s' % sam
    cmd += ' --seed 12345'
    cmd += ' --very-sensitive'
    cmd += ' -k %s' % params['k']
    cmd += ' --np %s' % params['np']
    cmd += ' --mp "%s"' % params['mp']
    cmd += ' --rdg "%s"' % params['rdg']
    cmd += ' --rfg "%s"' % params['rfg']
    cmd += ' --score-min "%s"' % params['score-min']
    cmd += ' --met-file %s.met' % splitext(sam)[0]
    cmd += ' --met 240'
    cmd += ' --no-unal'
    return cmd, sam


def get_alignment_cmd(
        fastxs: list,
        cmd: str,
        ali: str
) -> str:
    """Get the alignment command with ot without unix checks to decide
    whether the alignment (if exists) was aligned to completion, and not
    aborted (e.g., for lack of memory reasons).

    Parameters
    ----------
    fastxs : list
        Path to the input fasta/fastq/fastq.gz files
    cmd : str
        Alignment command
    ali : str
        Path to the alignment file

    Returns
    -------
    cmd : str
        Alignment command potentially decorated with unix check to rerun
    """
    if to_do(ali):
        return cmd
    else:
        return condition_ali_command(fastxs, cmd, ali)


def bowtie2(self) -> None:
    """Get the full command lines to perform sequence alignment using
     bowtie2 and potentially, re-run if the output was existing but
     possibly aborted.

    References
    ----------
    Langmead, B., & Salzberg, S. L. (2012). Fast gapped-read alignment with
    Bowtie 2. Nature methods, 9(4), 357-359.

    Notes
    -----
    GitHub  : https://github.com/BenLangmead/bowtie2
    Docs    : http://bowtie-bio.sourceforge.net/bowtie2/index.shtml
    Paper   : https://doi.org/10.1038/nmeth.1923

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for humann
        .sam_pool : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters for humann
        .config
            Configurations
    """
    for (tech, sample), fastxs in self.inputs[self.sam_pool].items():
        if tech_specificity(self, fastxs, tech, sample):
            continue
        to_dos = status_update(self, tech, fastxs)
        out = '%s/%s/%s' % (self.dir, tech, self.sam_pool)
        self.outputs['outs'][(tech, self.sam_pool)] = dict()
        for db, db_path in self.soft.params['databases'].items():
            db_out = '%s/%s/%s' % (out, db, self.soft.params['pairing'])
            cmd, sam = get_bowtie2_cmd(self, tech, fastxs, db_path, db_out)
            self.outputs['outs'][(tech, self.sam_pool)][(db, 'bowtie2')] = sam
            if self.config.force or to_do(sam):
                cmd = get_alignment_cmd(fastxs, cmd, sam)
                if to_dos:
                    self.outputs['cmds'].setdefault((tech,), []).append(False)
                else:
                    self.outputs['cmds'].setdefault((tech,), []).append(cmd)
                io_update(self, i_f=fastxs, i_d=db_out, o_d=db_out, key=tech)
                self.soft.add_status(tech, self.sam_pool, 1)
            else:
                self.soft.add_status(tech, self.sam_pool, 0)
            self.outputs['dirs'].append(db_out)


def salmon(self) -> None:
    """Salmon is a tool for wicked-fast transcript quantification from
    RNA-seq data. It requires a set of target transcripts (either from a
    reference or de-novo assembly) to quantify. All you need to run Salmon
    is a FASTA file containing your reference transcripts and a (set of)
    FASTA/FASTQ file(s) containing your reads. Optionally, Salmon can make
    use of pre-computed alignments (in the form of a SAM/BAM file) to the
    transcripts rather than the raw reads.

    References
    ----------
    Patro, Rob, et al. "Salmon provides fast and bias-aware quantification of
    transcript expression." Nature methods 14.4 (2017): 417-419.

    Notes
    -----
    GitHub  : https://github.com/COMBINE-lab/salmon
    Docs    : https://salmon.readthedocs.io/en/latest/salmon.html
    Paper   : https://doi.org/10.1038/nmeth.4197

    Parameters
    ----------
    self
    """
    print()


# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import click

from metagenomix.merger import merger
from metagenomix import __version__


@click.command()
@click.option("-i", "--folder", required=True,
              help='Path to pipeline output folder (`-o` for "create" module)')
@click.option("-o", "--summary",
              help='Output summary filename (will be in "<`-i`>/_managemnent")')
@click.option("-p", "--pipeline",
              help="Path to the file containing the softwares to run in order")
@click.option("-s", "--software", multiple=True,
              help="Software(s) to manage (or all in `-i/-p`)")
@click.option("--remove/--no-remove", default=True,
              help="Enable assistance to manage folder(s)/file(s) removals")
@click.option("--jobs/--no-jobs", default=True,
              help="Enable assistance to manage job stdour/stderr files")
@click.option("--rename/--no-rename", default=True,
              help="Enable assistance to manage folder(s)/file(s) renaming")
@click.option("--rename/--no-rename", default=True,
              help="Enable assistance to manage folder(s)/file(s) renaming")
@click.option("--verbose/--no-verbose", default=True, show_default=True,
              help="Whether to show input/outputs and other details")
@click.version_option(__version__, prog_name="metagenomix")
def merge(
        folder,
        summary,
        pipeline,
        software,
        remove,
        jobs,
        rename,
        verbose
):
    """Combine the per-sample outputs into feature tables."""
    merger(
        folder=folder,
        out=summary,
        pipeline=pipeline,
        softwares=software,
        remove=remove,
        jobs=jobs,
        rename=rename,
        verbose=verbose
    )



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import click

from metagenomix.manager import manager
from metagenomix import __version__


@click.command()
@click.option("-i", "--folder", required=True,
              help='Path to pipeline output folder (the folder passed to `-o` '
                   'for "metagenomix create")')
@click.option("-o", "--storage", default='', required=True,
              help="Path to pipeline output folder on storage disk "
                   "(for `--store` task)")
@click.option("-p", "--pipeline", default=None, show_default=True,
              help="Path to the file containing the softwares to run in order")
@click.option("-s", "--software", multiple=True,
              help="Software(s) to manage (or all in `-i/-p`)")
@click.option("-x", "--chunks", type=int, show_default=False, default=None,
              help="Number of scripts for each `--store` task")
@click.option("--jobs/--no-jobs", show_default=True,
              help="[Task] Enable job output management")
@click.option("--remove/--no-remove", default=False, show_default=True,
              help="[Task] Enable output removal")
@click.option("--rename/--no-rename", default=False, show_default=True,
              help="[Task] Enable output renaming")
@click.option("--store/--no-store", default=False, show_default=True,
              help="[Task] Enable output storage")
@click.option("--confirm/--no-confirm", default=True, show_default=True,
              help="Whether to ask for confirmation before applying task")
@click.version_option(__version__, prog_name="metagenomix")
def manage(
        folder,
        storage,
        pipeline,
        software,
        chunks,
        jobs,
        remove,
        rename,
        store,
        confirm
):
    """Deal with the contents of your pipeline output folder."""
    manager(
        dir=folder,
        disk=storage,
        pipeline=pipeline,
        softwares=software,
        chunks=chunks,
        jobs=jobs,
        remove=remove,
        rename=rename,
        store=store,
        confirm=confirm
    )



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import re
import sys
import yaml
import glob
import hashlib
import itertools
import pandas as pd
from tabulate import tabulate
from os.path import basename, dirname, isdir, isfile, islink


def read_yaml(
        file_path: str
) -> dict:
    """Simply reads a yaml and return its contents in a dictionary structure.

    Parameters
    ----------
    file_path: str
        Path to a yaml file.

    Returns
    -------
    yaml_dict : dict
        Dictionary returned by reading the yaml file (could be empty).
    """
    yaml_dict = {}
    if file_path:
        if not isfile(file_path):
            raise IOError('No yaml file "%s"' % file_path)
        with open(file_path) as yaml_handle:
            try:
                yaml_dict = yaml.load(yaml_handle, Loader=yaml.FullLoader)
            except AttributeError:
                yaml_dict = yaml.load(yaml_handle)
    return yaml_dict


def get_fastq_files(
        fastqs: list
) -> list:
    fastq_gz = [fastq for fastq in fastqs if '.gz' in fastq]
    if len(fastq_gz):
        return fastq_gz
    else:
        return fastqs


def get_fastq_paths(
        fastq_dirs: list
) -> list:
    fastqs = []
    for fastq_dir in fastq_dirs:
        fastqs.extend(glob.glob(fastq_dir + '/*.fastq*'))
    return fastqs


def mkdr(
        path: str,
        is_file: bool = False
) -> None:
    """Creates a folder is does not exist yet.

    Parameters
    ----------
    path : str
        Path of the file or folder to create.
    is_file : bool
        True is the path is a file path, False is it is a folder path.
    """
    if is_file:
        os.makedirs(dirname(path), exist_ok=True)
    else:
        os.makedirs(path, exist_ok=True)


def get_pfam_wget_cmd(
        pfam_dir: str
) -> str:
    """Get the validation (even through downloading) that the HMM file is there.

    Parameters
    ----------
    pfam_dir : str
        Path to the folder that should contain the "Pfam-A.*" files

    Returns
    -------
    cmd : str
        Command line to run to download the database
    """
    cmd = ''
    for ext in ['hmm', 'fasta', 'hmm.dat']:
        fp = '%s/Pfam-A.%s' % (pfam_dir, ext)
        if not isfile(fp):
            gz = 'pub/databases/Pfam/releases/Pfam35.0/Pfam-A.%s.gz' % ext
            cmd += 'wget -O %s http://ftp.ebi.ac.uk/%s\n' % (fp, gz)
            cmd += 'gunzip %s\n' % fp
    return cmd


def get_hmm_dat(
        pfam_dir: str
) -> pd.DataFrame:
    """Get as table the contents of
    http://ftp.ebi.ac.uk/pub/databases/Pfam/releases/Pfam35.0/Pfam-A.hmm.dat.gz

    Parameters
    ----------
    pfam_dir : str
        Path to the folder that should contain the "Pfam-A.*" files

    Returns
    -------
    pfam_pd : pd.DataFrame
        Pfam dat reformatted as table.
    """
    tsv = '%s/Pfam-A.hmm.dat.tsv' % pfam_dir
    if not isfile(tsv):
        records = []
        cur = {}
        with open('%s/Pfam-A.hmm.dat' % pfam_dir) as f:
            for line in f:
                line_decode = line.strip()
                if line_decode.startswith('//'):
                    records.append(cur)
                    cur = {}
                elif line_decode.startswith('#=GF'):
                    cur[line_decode[5:7]] = line_decode[7:].strip()
        pfam_pd = pd.DataFrame(records)
        pfam_pd.to_csv(tsv, index=False, sep='\t')
    else:
        pfam_pd = pd.read_table(tsv)
    return pfam_pd


def get_hmms_dias_cmd(
        hmm: str,
        term_pd: pd.DataFrame,
        term: str,
        odir: str
) -> tuple:
    """Collect the hmm profiles and diamond database for different term targets.

    Parameters
    ----------
    hmm : str
        Pfam-A.hmm file.
    term_pd : pd.DataFrame
        Data frame containing the Pfam-A.hmm.dat file, but reformatted above.
    odir : str
        Directory where all Pfam models will be kept.
    term : str
        Term to search among the Pfam hmms.

    Returns
    -------
    hmms_dias : dict
        [.hmm, .dmnd] files for each term.
    cmd : str
        Command to hmmfetch these term's hmms.
    """
    comp = re.compile("[\\ \-\",()%':&/.\[\]]")
    cmd = ''
    hmms_dias = {}
    pfam_out = '%s/%s' % (odir, comp.sub('_', term))
    mkdr(pfam_out)
    for r, row in term_pd.iterrows():
        acc, desc = row['AC'], comp.sub('_', row['DE'])
        name = '%s__%s' % (acc, desc)
        hmm_fp = '%s/%s.hmm' % (pfam_out, name)
        if not isfile(hmm_fp):
            cmd += 'hmmfetch %s %s >> %s\n' % (hmm, acc, hmm_fp)
        fa = '%s/fastas/%s.fa' % (odir, acc)
        fo = '%s/%s.fa' % (pfam_out, name)
        dia = '%s/%s.dmnd' % (pfam_out, name)
        if not isfile(dia):
            cmd += 'cp %s %s\n' % (fa, fo)
            cmd += 'diamond makedb --in %s -d %s\n' % (fo, dia)
            cmd += 'rm %s\n' % fo
        hmms_dias[name] = [hmm_fp, dia]
    return hmms_dias, cmd


def min_nlines(
        input_fp: str
) -> bool:
    """Check whether the number of lines in the file is above one.

    Parameters
    ----------
    input_fp : str
        Path to the input file.

    Returns
    -------
    ret : bool
        Whether the file contains at least one sequence.
    """
    ret = False
    with open(input_fp.replace('${SCRATCH_FOLDER}', '')) as f:
        for ldx, line in enumerate(f):
            if ldx > 1:
                ret = True
                break
    return ret


def not_paired(
        self,
        tech: str,
        sam: str,
        fqs: list,
) -> bool:
    """Checks whether there are two input files, which is what is needed for
    merging. Otherwise, stop there and show a useful error message.

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    sam : str
        Name of the current sample
    fqs : list
        Paths to the input files

    Returns
    -------
    bool
        Whether the input files are not possibly pooled or not
    """
    nfiles = len(fqs)
    if nfiles != 2:
        self.outputs['outs'].setdefault((tech, self.sam_pool), []).extend(fqs)
        self.soft.add_status(tech, sam, 1, message='unpaired reads')
        return True
    return False


def status_update(
        self,
        tech: str,
        inputs: list,
        pool: str = None,
        group: str = None,
        genome: str = None,
        message: str = None,
        folder: bool = False,
        software: str = ''
) -> list:
    """Potentially add the fastq files to the status (files to generate).

    Parameters
    ----------
    self : Commands class instance
        .sam_pool : str
            Sample name
        .soft
            Software class instance
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    inputs : list
        Paths to input files
    pool : str
        Name of the current co-assembly
    group : str
        Name of the current co-assembly group
    genome : str
        MAGs/Genomes folder name or empty string (for assembly contigs)
    message : str
        Warning to show
    folder : bool
        Whether the paths are folders
    software : str
        Software name to allow collecting status

    Returns
    -------
    to_dos : list
        Paths to input file that need to be created (or whether they are links)
    """
    to_dos = get_to_dos(self, inputs, folder)
    if to_dos and self.soft.name != software:
        if pool:
            self.soft.add_status(tech, pool, to_dos,
                                 group=group, genome=genome, message=message)
        else:
            self.soft.add_status(tech, self.sam_pool, to_dos,
                                 group=group, genome=genome, message=message)
    return to_dos


def get_to_dos(
        self,
        inputs: list,
        folder: bool
) -> list:
    """Potentially add the fastq files to the status (files to generate).

    Parameters
    ----------
    self : Commands class instance
        .sam_pool : str
            Sample name
        .soft
            Software class instance
    inputs : list
        Paths to input files
    folder : bool
        Whether the paths are folders

    Returns
    -------
    to_dos : list
        Paths to input file that need to be created (or whether they are links)
    """
    to_dos = []
    for inp in inputs:
        if folder:
            to_do_inp = to_do(folder=inp)
        else:
            to_do_inp = to_do(inp)
        if to_do_inp:
            to_dos.append(inp)
        if isinstance(inp, list):
            inps = inp
        else:
            inps = [inp]
        inps = [x.replace('${SCRATCH_FOLDER}', '') for x in inps]
        self.soft.links.update({x: os.readlink(x) for x in inps if islink(x)})
    return to_dos


def get_roundtrip(io) -> dict:
    roundtrip = {'to': inputs_to_scratch(io), 'from': outputs_back(io)}
    return roundtrip


def inputs_to_scratch(io) -> list:
    rsyncs, mkdirs = set(), set()
    # folders
    if ('I', 'd') in io:
        for folder_ in io[('I', 'd')]:
            folder = folder_.rstrip('/')
            src = folder_.rstrip('/').replace('${SCRATCH_FOLDER}', '')
            mkdirs.add('mkdir -p %s' % folder)
            rsyncs.add('rsync -aqruv %s/ %s' % (src, folder))
    # folders
    if ('O', 'd') in io:
        for folder in io[('O', 'd')]:
            mkdirs.add('mkdir -p %s' % folder.rstrip('/'))
    # files
    if ('I', 'f') in io:
        for file in io[('I', 'f')]:
            folder = dirname(file)
            src = file.replace('${SCRATCH_FOLDER}', '')
            mkdirs.add('mkdir -p %s' % folder)
            rsyncs.add('rsync -aqruv %s %s' % (src, file))
    return sorted(mkdirs) + sorted(rsyncs)


def outputs_back(io) -> list:
    outbound = set()
    if ('O', 'd') in io:
        # folders
        for folder_ in io[('O', 'd')]:
            folder = folder_.rstrip('/')
            src = folder_.rstrip('/').replace('${SCRATCH_FOLDER}', '')
            cmd = 'mkdir -p %s; rsync -aqruv %s/ %s' % (src, folder, src)
            cmd = 'if [ -d %s ]; then %s; fi' % (folder, cmd)
            outbound.add(cmd)
    if ('O', 'f') in io:
        # files
        for file in io[('O', 'f')]:
            src = file.replace('${SCRATCH_FOLDER}', '')
            folder = dirname(src)
            cmd = 'mkdir -p %s; rsync -aqruv %s %s' % (folder, file, src)
            cmd = 'if [ -f %s ]; then %s; fi' % (file, cmd)
            outbound.add(cmd)
    return sorted(outbound)


def get_scratch_cmds(
        self,
        key,
        soft,
        cur_cmds,
        cmds
) -> dict:
    if key in soft.io:
        roundtrip = get_roundtrip(soft.io[key])
        scratch_cmds = ['\n# Move to SCRATCH_FOLDER'] + roundtrip['to']
        scratch_cmds += ['\n# %s commands (%s)' % (soft.name, key)] + cur_cmds
        if self.config['move_back']:
            scratch_cmds += ['\n# Move from SCRATCH_FOLDER'] + roundtrip['from']
        cmds[key] = scratch_cmds
    else:
        cmds[key] = cur_cmds


def per_group_scratch(
        self,
        pool,
        soft,
        sam_cmds,
        cmds,
        commands
) -> None:
    if pool in sam_cmds:
        get_scratch_cmds(self, pool, soft, sam_cmds, cmds)
    else:
        for group in commands.pools[pool]:
            group_cmds = sam_cmds[group]
            get_scratch_cmds(self, (pool, group), soft, group_cmds, cmds)


def scratching(
        self,
        soft,
        commands
) -> dict:
    if soft.params['scratch'] and self.config.jobs:
        cmds = {}
        # Use commands.pools to unpack the commands well
        for sam, sam_cmds in soft.cmds.items():
            if isinstance(sam_cmds, list):
                get_scratch_cmds(self, sam, soft, sam_cmds, cmds)
            elif isinstance(sam_cmds, dict):
                per_group_scratch(self, sam, soft, sam_cmds, cmds, commands)
            else:
                sys.exit('The collected commands are neither list of dict!')
        return cmds
    else:
        return soft.cmds


def io_update(
        self,
        i_f=None,
        i_d=None,
        o_f=None,
        o_d=None,
        key=None
):
    if not isinstance(key, tuple):
        key = (key,)
    for (IO_fd, val) in [
        (('I', 'f'), i_f),
        (('I', 'd'), i_d),
        (('O', 'f'), o_f),
        (('O', 'd'), o_d)
    ]:
        if not val:
            continue
        if isinstance(val, list):
            self.outputs['io'][IO_fd].setdefault(key, set()).update(val)
        elif isinstance(val, str):
            self.outputs['io'][IO_fd].setdefault(key, set()).add(val)


def to_do(
        file: str = None,
        folder: str = None
) -> bool:
    if file:
        file = file.replace('${SCRATCH_FOLDER}', '')
        if isfile(file) or islink(file):
            return False
    if folder:
        folder = folder.replace('${SCRATCH_FOLDER}', '')
        if (isdir(folder) or islink(folder)) and glob.glob('%s/*' % folder):
            return False
    return True


def tech_specificity(
        self,
        data,
        tech: str,
        sam: str,
        specificity: list = []
) -> bool:
    """Returns a boolean that is True if the current technology can not be
    processed by the current tool, possibly because is has no input files.
    If there are files but the technology can not be processed, the files
    are passed as output so that they can be available to the next steps.

    Notes
    -----
    This behaviour is common to all software functions and therefore
    justifies this function, and some softwares of a workflow ar not designed
    for some technologies so the pipeline should pass through these while
    retaining the files available for a next step.
    For example, the reads merging can only happen on the 'illumina' data.
    So the 'pacbio' and 'nanopore' data will just be passed as output to read
    merging (without merging) so that they can be available to the next step.

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    sam : str
        Sample name
    data : str
        Path to the input files or other input data structure
    specificity : list
        Technology that can be processed by the current software

    Returns
    -------
    bool
        Whether the technology is possibly processed using this tool
    """
    if not data or (specificity and tech not in specificity):
        self.outputs['outs'].setdefault((tech, sam), []).extend(data)
        self.soft.add_status(tech, sam, 0, message='technology incompatible')
        return True
    return False


def caller(
        self,
        namespace
):
    """Calls as function the part of the software name that follows the first
    underscore.
    For example, software name "search_diamond" would call function`diamond()`.
    """
    module = sys.modules[namespace]
    for sub in ['', '_']:
        func = self.soft.name.split('_', 1)[-1] + sub
        if func == self.soft.name:
            continue
        if hasattr(module, func) and callable(getattr(module, func)):
            module_call = getattr(module, func)
            return module_call
    print('No function "%s" in module "%s"' % (func, namespace))


def edit_sample_or_pool(not_done):
    sample_or_pool = 'pool'
    if not_done['group'].isnull().all():
        sample_or_pool = 'sample'
    not_done.rename(columns={'sample_or_pool': sample_or_pool}, inplace=True)


def show_not_done(not_done):
    """

    Parameters
    ----------
    not_done : pd.DataFrame

    Returns
    -------
    status_summary : str
        Summary of the analyses that remain to do
    """
    # replace "sample_or_pool" by "sample" or "pool" depending on the data type
    edit_sample_or_pool(not_done)
    # get the data that remains to analyse for the current software
    todo_pd = not_done.loc[not_done['status'] == 'To do'].copy()
    # make a one-liner summarizing what remains to analyse (per tech, pool etc)
    status_summary = summarize_status(todo_pd)
    print(status_summary)
    return status_summary


def print_vals(vals) -> str:
    vals = '; '.join(['%ss' % v if int(v.split()[0]) > 1 else v for v in
                        vals])
    return vals


def summarize_status(tab) -> str:
    """

    Parameters
    ----------
    tab

    Returns
    -------
    status_summary : str
        Summary of the analyses that remain to do
    """
    cols = tab.columns.tolist()[:2] + ['group', 'genome']
    dat = tab[cols].apply(lambda x: x.nunique()).astype(str).to_dict().items()
    vals = [' '.join(x[::-1]) for x in dat if int(x[1])]
    if vals:
        status_summary = 'to do:\t%s' % print_vals(vals)
    else:
        status_summary = 'nothing to do'
    return status_summary


def pretty_print(tab, header, typ) -> str:
    # stackoverflow.com/questions/41593793/
    # printtabulate-to-pretty-print-multiindex-pandas
    p_tab = tabulate(tab, headers=header, tablefmt='orgtbl')
    tab_width = [len(x) for x in p_tab.split('\n') if '|--' in x][0]
    dash = '-' * tab_width
    if not typ:
        typ = 'needed input'
    gap = (tab_width // 2) - (len(typ) // 2)
    header = '%s\n%s%s\n%s\n' % (dash, ' ' * gap, typ, dash)
    out_tab = header + p_tab + '\n' + dash
    p_tab = '\n\t\t' + re.sub('\\n', '\n%s' % ('\t' * 2), out_tab) + '\n'
    print(p_tab)
    return out_tab


def get_pivot_table_cols(
        tab: pd.DataFrame,
) -> list:
    """Get the columns of the status table that are to be used
    to report progress.

    Parameters
    ----------
    tab : pd.DataFrame

    Returns
    -------
    pivot_table_cols : list
    """
    tab_cols = tab.columns.tolist()
    cols = tab_cols[:2] + ['group']
    if 'genome' in tab_cols:
        cols += ['genome']
    pivot_table_cols = [col for col in cols if not tab[col].isnull().all()]
    return pivot_table_cols


def pivot_not_done_multi_header(t) -> tuple:
    # reorder column multiindex to get the tech above and group/genome below
    t.columns = t.columns.reorder_levels(range(len(t.columns.levels))[::-1])
    # sort table columns to suit the reordering
    t = t[[
        it for it in itertools.product(*[
            lev[::-1] if lx else lev for lx, lev in enumerate(t.columns.levels)
        ])
    ]]
    # edit the multi-index lie for pretty printing
    h = [t.index.names[0] + '/' + t.columns.names[0]] + list(
        map('\n'.join, t.columns.tolist()))
    return t, h


def print_tab_pv(
        tab_pv: pd.DataFrame,
        cols: list,
        typ: str = '',
) -> tuple:
    """

    Parameters
    ----------
    tab_pv : pd.DataFrame
    cols : list
    typ : str

    Returns
    -------
    out_tab : str
        Table to write out
    skip : bool
        Whether the print table is only for sample per tech
        (will break without printing content details since there is no content)
    """
    skip = False
    if len(cols[2:]) > 1:
        tab_pv_sorted, h = pivot_not_done_multi_header(tab_pv)
        out_tab = pretty_print(tab_pv_sorted, h, typ)
    elif len(cols[2:]):
        h = [tab_pv.index.names[0]] + ['\n'.join(x[::-1]) if x[0][:3] == 'Run'
                                       else x[-1] for x in tab_pv.columns]
        out_tab = pretty_print(tab_pv, h, typ)
    else:
        h = [tab_pv.index.names[0]] + tab_pv.columns.tolist()
        out_tab = pretty_print(tab_pv, h, typ)
        skip = True
    return out_tab, skip


def print_message(
        tab: pd.DataFrame,
) -> list:
    """Loop to print both the data remaining to analyze, both in terms of
    job counts and job contents (sample, co-assembly pool's group, genome).

    Parameters
    ----------
    tab : pd.DataFrame
        Status table for the

    Returns
    -------
    out_messages : list
        Pretty tables
    """
    out_messages = []
    cols = get_pivot_table_cols(tab)
    for (typ, aggfunc) in [
        ('counts', 'count'),
        ('contents', (lambda x: '\n'.join([y for y in x if y is not None]))),
    ]:
        tab_pv = tab[cols].pivot_table(
            values=cols[2:],
            columns=cols[0],
            index=cols[1],
            aggfunc=aggfunc)
        out_tab, break_here = print_tab_pv(tab_pv, cols, typ)
        out_messages.append(out_tab)
        if break_here:
            break
    return out_messages


def print_messages(not_done):
    out_messages = []
    for mess, mess_pd in not_done.groupby('message'):
        message = 'Attention -> "%s":' % mess
        print('\n\t\t%s ' % message)
        out_messages.append(message)
        out_messages.extend(print_message(mess_pd))
    return out_messages


def print_needed(path, not_done) -> str:
    tab = not_done.loc[not_done['status'] != 'To do'].copy()
    if len(tab):
        tab['status'] = ['\n'.join(map(basename, x)) for x in tab['status']]
        path = ['%s> %s' % (' ' * x, y) for x, y in enumerate(path[:-1])]
        status = 'Run:\n%s\n%s*' % ('\n'.join(path[:-1])[2:],
                                    path[-1].replace('> ', '=> *'))
        tab = tab.rename(columns={'status': status})
        cols = get_pivot_table_cols(tab) + [status]
        tab_pv = tab[cols].pivot_table(
            values=cols[2:],
            columns=cols[0],
            index=cols[1],
            aggfunc=(lambda x: '\n'.join(set(y for y in x if y is not None))))
        out_tab, _ = print_tab_pv(tab_pv, cols)
        return out_tab


def print_status_table(
        soft,
        show_status: bool = False
) -> None:
    if soft.status:
        status = pd.DataFrame(soft.status, columns=[
            'tech', 'sample_or_pool', 'status', 'group', 'message', 'genome'])
        status.drop_duplicates(inplace=True)
        not_done = status.loc[status['status'] != 'Done'].copy()
        if len(not_done):
            soft.tables.append(show_not_done(not_done))
            if show_status:
                # pretty table of data that remains to analyse when message
                soft.tables.extend(print_messages(not_done))
                # pretty table of data that remains to analyse upfront
                soft.tables.append(print_needed(soft.path, not_done))
        else:
            soft.tables.append('done')
            print('done')
    else:
        soft.tables.append('done')
        print('done')


def compute_hash(hash_string):
    h = hashlib.blake2b(digest_size=10)
    h.update(str(hash_string).encode('utf-8'))
    hashed = str(h.hexdigest())
    return hashed


def get_md5(fp):
    if isdir(fp):
        return ''
    md5 = hashlib.md5()
    with open(fp, "rb") as f:
        for chunk in f:
            md5.update(chunk)
    md5 = str(md5.hexdigest())
    return md5


def get_dates(run_fp) -> list:
    """Get the different dates at which the current pipeline configuration
    was run.

    Parameters
    ----------
    run_fp
        Path to the versioning file for this pipeline configuration
    Returns
    -------
    runs : list
        All the dates at which this pipeline configuration was run
    """
    runs = []
    if isfile(run_fp):
        with open(run_fp) as f:
            for line in f:
                if line.startswith('Date'):
                    runs.append(line.strip().split()[-1])
    return runs


def get_input_info(dat):
    m = '%s input unit' % dat.shape[0]
    if dat.shape[0] > 1:
        m += 's'
    info = '%s scripts (%s):' % (dat.name.nunique(), m)
    return info


def get_res_info(dat):
    d = [(len(v), [len(x) for x in v.values()]) for v in dat.values() if v]
    if not d:
        return 'no data'
    tech = '%s tech' % len(d)
    if len(d) > 1:
        tech += 's'
    fs = sum([x[0] for x in d])
    folders = '%s folder' % fs
    if fs > 1:
        folders += 's'
    av = round(sum([round(sum(x[1])/len(x[1]), 2) for x in d]) / len(d), 2)
    files = '%s file' % av
    if av > 1:
        files += 's'
    info = '%s (%s and %s per folder)' % (tech, folders, files)
    return info


def get_size_info(dat):
    sizes = {}
    for folder, size_ in dat.items():
        if size_ < 1024:
            size = "%s bytes" % size_
        elif size_ < 1024 * 1024:
            size = "%s KB" % round(size_ / 1024, 2)
        elif size_ < 1024 * 1024 * 1024:
            size = "%s MB" % round(size_ / (1024 * 1024), 2)
        elif size_ < 1024 * 1024 * 1024 * 1024:
            size = "%s GB" % round(size_ / (1024 * 1024 * 1024), 2)
        else:
            size = str(size_)
        if size_ > 0:
            sizes[folder] = size
    fs = '%s folder' % len(sizes)
    if len(sizes) > 1:
        fs += 's'
    info = '%s (./%s)' % (fs, '; ./'.join(['='.join(x) for x in sizes.items()]))
    return info, sizes



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import re
import os
from glob import glob
import pandas as pd
import pkg_resources
from os.path import basename, getsize, isdir, isfile, splitext


class Output(object):

    def __init__(self, folder, soft):
        self.soft = soft
        self.soft_dir = folder + '/' + soft
        # summary information about
        self.name = None
        self.after = None
        self.script = None
        self.after_dir = None
        self.directives = {'account', 'partition', 'gres',
                           'time', 'mem', 'ntasks', 'job-name'}
        self.oe = []
        self.hpc = []
        self.input = []
        self.outputs = {}
        self.sample_pool = {}
        self.run_scripts = {}
        self.dir_scripts = {}

    def get_afters(self):
        for folder in os.listdir(self.soft_dir):
            if folder.startswith('after_'):
                key = tuple(folder.split('after_')[-1].rsplit('_', 1))
                self.outputs[key] = {'stdouts': {}, 'results': {}, 'sizes': {}}

    def get_outputs(self):
        for after in list(self.outputs):
            self.after = after
            self.get_contents()
            self.get_results()
            self.parse_provenance()
            self.parse_run_sh()
            self.job_outputs()
            self.make_table(after)

    def get_contents(self):
        self.after_dir = '%s/after_%s' % (self.soft_dir, '_'.join(self.after))
        all_scripts = set(glob('%s/jobs/*.*' % self.after_dir))
        self.outputs[self.after]['all_scripts'] = all_scripts

    def get_results(self):
        for data_dir in [x for x in os.listdir(self.after_dir) if x != 'jobs']:
            cur_dir = self.after_dir + '/' + data_dir
            if not isdir(cur_dir):
                continue
            self.outputs[self.after]['results'][data_dir] = {}
            self.outputs[self.after]['sizes'][data_dir] = 0
            for root_, dirs, fs in os.walk(cur_dir):
                if root_ == cur_dir:
                    continue
                root = root_.split('%s/' % cur_dir)[-1]
                sizes = sum([getsize('%s/%s' % (root_, x)) for x in fs])
                self.outputs[self.after]['sizes'][data_dir] += sizes
                self.outputs[self.after]['results'][data_dir].update({root: fs})

    def parse_provenance(self):
        self.sample_pool[self.after] = 'sample'
        provenance = '%s/provenance.txt' % self.after_dir
        if isfile(provenance):
            with open(provenance) as provenance_handle:
                for line in provenance_handle:
                    if not line.strip():
                        break
                    if line.split()[1] == 'pooling':
                        self.sample_pool[self.after] = 'pool'
                        break

    def parse_run_sh(self):
        run_sh = '%s/run.sh' % self.after_dir
        if isfile(run_sh):
            scripts = set()
            with open(run_sh) as run_handle:
                for ldx, line in enumerate(run_handle):
                    if ldx < 2:
                        continue
                    self.script = line.strip().split()[-1]
                    scripts.add(self.script)
                    if self.script.endswith('.slm'):
                        self.parse_slm()
                        self.job_outputs(self.name)
                        self.outputs[self.after][self.name] = self.script
                    self.get_inputs()
            self.outputs[self.after]['run_scripts'] = scripts

    def make_table(self, after):
        for (tab, attr) in [
            ('oe', self.oe),
            ('hpc', self.hpc),
            ('input', self.input),
        ]:
            self.outputs[after][tab] = pd.DataFrame(attr)
        self.oe = []
        self.hpc = []
        self.input = []

    def parse_slm(self):
        directives = {}
        with open(self.script) as f:
            for line in f:
                if line.startswith('#SBATCH') and '=' in line:
                    key, value = line.strip().split('SBATCH ')[-1].split('=')
                    if key.strip('-') in self.directives:
                        directives[key.split('-')[-1]] = value
        self.name = directives.get('name')
        self.hpc.append(directives)

    def job_outputs(self, job_name: str = ''):
        o_fps = '%s/jobs/output/*%s*.o' % (self.after_dir, job_name)
        for sdx, stdout in enumerate(sorted(glob(o_fps))):
            name, job_id = basename(stdout[:-2]).split('-', 1)[1].rsplit('_', 1)
            self.outputs[self.after]['stdouts'][(name, job_id)] = stdout

            status = {'name': name, 'done': 'N', 'id': job_id, 'script': None,
                      'cpu_usage': None, 'memory_usage': None, 'error': None}
            ls = open(stdout).readlines()
            status['script'] = [x for x in ls[:15]][0].strip().split('/')[-1]
            if 'Done!' in ls:
                status['done'] = 'Y'
            if 'Task and CPU usage stats:\n' in ls:
                line = ls[ls.index('Task and CPU usage stats:\n') + 4].split()
                status['cpu_usage'] = ' '.join([line[2], line[6]])
            if 'Memory usage stats:\n' in ls:
                line = ls[ls.index('Memory usage stats:\n') + 4].split()
                status['memory_usage'] = line[3]

            stderr = '%s.e' % splitext(stdout)[0]
            ls = open(stderr).readlines()
            if ls and 'error' in ls[-1]:
                status['error'] = ls[-1]
            self.oe.append(status)

    def get_inputs(self):
        cols = [self.sample_pool[self.after], 'tech', 'group', 'genome_taxon']
        if isfile(self.script):
            with open(self.script) as script_handle:
                for line in script_handle:
                    if line.startswith('# data: '):
                        data = line[8:].strip().split()
                        if len(data) < 4:
                            data += [None] * (4 - len(data))
                        data = {col: data[x] for x, col in enumerate(cols)}
                        data['name'] = self.name
                        self.input.append(data)


class Softwares(object):

    def __init__(self, **kwargs) -> None:
        self.__dict__.update(kwargs)
        self.inputs = {'res': {}, 'dir': set(), 'pip': set(), 'usr': set()}
        self.names = set()
        self.roles = {}
        self.softs = {}
        self.get_softs()

    def _from_dir(self):
        self.inputs['dir'].update([soft for soft in os.listdir(self.dir)])

    def _from_usr(self):
        for soft in self.softwares:
            if soft in self.inputs['dir']:
                self.inputs['usr'].add(soft)
            else:
                r = re.compile(soft.replace('*', '.*').replace('..*', '.*'))
                self.inputs['usr'].update(
                    set(filter(r.match, list(self.inputs['dir']))))

    def _from_res(self):
        res = pkg_resources.resource_filename("metagenomix", "resources")
        with open('%s/softwares.txt' % res) as f:
            for line in f:
                soft, role = line.strip().split('\t')
                self.roles.setdefault(role.split(' (')[0], set()).add(soft)
                self.inputs['res'][soft] = role

    def _from_pipeline(self):
        if self.pipeline:
            with open(self.pipeline) as f:
                self.inputs['pip'].update([x.strip().split()[-1] for x in
                                           f if x.strip() and x[0] != '#'])

    def _intersection(self) -> set:
        softs = self.inputs['dir'] & set(self.inputs['res'])
        if self.inputs['pip'] | self.inputs['usr']:
            softs = softs & (self.inputs['pip'] | self.inputs['usr'])
        return softs

    def _to_manage(self) -> dict:
        softs_intersection = self._intersection()
        for role, softs in self.roles.items():
            cur_softs = softs & softs_intersection
            if cur_softs:
                self.softs[role] = {t: self.inputs['res'][t] for t in cur_softs}
                self.names.update(cur_softs)

    def show(self):
        print('Tools to %s per role:' % self.command)
        for role, softs in self.softs.items():
            print('  [Role: %s]' % role)
            n = max([len(soft) for soft in softs])
            roles = set(softs.values())
            for soft, role_ in softs.items():
                print('\t- %s' % soft, end='')
                if role_ == role or len(roles) == 1:
                    print()
                else:
                    print(' ' * (n-len(soft)), '\t:', role_.split('(')[-1][:-1])

    def get_softs(self):
        print('* Getting softwares to %s' % self.command)
        self._from_dir()
        self._from_usr()
        self._from_res()
        self._from_pipeline()
        self._to_manage()
        self.show()



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from metagenomix.metagenomix import metagenomix
from metagenomix.core.jobs import Created


def creator(**kwargs):
    """Run metagenomix to account for the passed command-line arguments.

    Parameters
    ----------
    kwargs : dict
        All arguments passed in command line, including defaults
    """
    print('\n>>> `metagenomix create` started >>>\n')
    # Collect all command and init the script creating instance
    creating = Created(*metagenomix(**kwargs))
    # print('* Writing database formatting commands')
    # scripting.database_cmds()  # build the databases
    print('* Creating output folders')
    creating.make_dirs()
    print('* Writing pipeline command lines')
    creating.software_cmds()   # run the analysis pipeline
    if len(creating.run['database']) or len(creating.run['software']):
        m = '\n< PLEASE CONSIDER CHECKING THE COMMAND LINE SCRIPTS MANUALLY >'
        print(m)
        creating.display()  # show the scripts to run
    creating.write_links()
    creating.write_logs()
    print('\n<<< `metagenomix create` completed <<<\n')



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import itertools
from os.path import abspath

from metagenomix._io_utils import compute_hash
from metagenomix._inputs import show_inputs
from metagenomix.softwares.alignment import *
from metagenomix.softwares.annotation import *
from metagenomix.softwares.args import *
from metagenomix.softwares.assembly import *
from metagenomix.softwares.binning import *
from metagenomix.softwares.metamarker import *
from metagenomix.softwares.genomics import *
from metagenomix.softwares.phlans import *
from metagenomix.softwares.plasmids import *
from metagenomix.softwares.pooling import pooling
from metagenomix.softwares.preprocess import *
from metagenomix.softwares.profiling import *
from metagenomix.softwares.simka import *
from metagenomix.softwares.strains import *
from metagenomix.softwares.viruses import *
# from metagenomix.softwares.mapping import mapping


class Commands(object):

    def __init__(self, config, databases, workflow):
        self.config = config
        self.databases = databases
        self.graph = workflow.graph
        self.softs = workflow.softs
        self.outputs = {}
        self.cmds = {}
        self.args = {}
        self.pools = {}
        self.longs = None
        self.inputs = None
        self.method = None
        self.soft = None
        self.sam_pool = None
        self.dir = ''
        self.out = []
        self.struc = list
        self.links = {}
        self.links_stats = {}
        self.holistics = [
            'antismash',
            'simka',
            'quast',
            'qiita_wol',
            'mag_data',
            'metamarker',
            'woltka',
            'drep',
            'strainphlan'
        ]

    def collect(self):
        for sdx, softs in enumerate(self.config.pipeline):
            print(' - %s' % softs[-1])
            self.soft = self.softs[softs[-1]]
            self.soft.add_soft_path(self.softs)
            self.get_inputs()
            self.get_hash()
            self.get_dir()
            self.generic_command()
            self.show_messages()

    def get_inputs(self):
        """Update the `inputs` attribute of the software object."""
        if not self.soft.prev or self.soft.name == 'map__drep':
            if self.soft.params['scratch'] and self.config.jobs:
                self.inputs = self.config.fastq_mv
            else:
                self.inputs = self.config.fastq
        else:
            self.inputs = self.softs[self.soft.prev].outputs
        show_inputs(self)

    def get_hash(self):
        avoid = {'time', 'nodes', 'mem', 'mem_dim', 'env', 'chunks',
                 'scratch', 'machine', 'partition'}
        params = dict(x for x in self.soft.params.items() if x[0] not in avoid)
        path = self.soft.path
        hashes = [self.softs[x].hash for x in self.soft.path[1:-1]]
        self.soft.hash = (params, path, hashes)
        self.soft.hashed = compute_hash(self.soft.hash)

    def get_dir(self):
        self.dir = abspath('%s/%s/after_%s_%s' % (
            self.config.dir, self.soft.name, self.soft.prev, self.soft.hashed))
        self.soft.dir = self.dir
        if self.soft.params['scratch'] and self.config.jobs:
            self.dir = '${SCRATCH_FOLDER}%s' % self.dir

    def is_pool(self):
        self.struc = list
        if set(self.inputs) == set(self.pools) or self.soft.name == 'pooling':
            self.struc = dict

    def generic_command(self):
        self.sam_pool = ''
        self.is_pool()
        self.soft.io = {}
        if self.soft.name in self.holistics:
            self.prep_job()
        elif self.soft.name == 'pooling':
            self.pooling()
        else:
            for sam_or_pool in sorted(self.inputs):
                self.sam_pool = sam_or_pool
                self.prep_job()
        self.register_command()

    def show_messages(self):
        for message in self.soft.messages:
            print('[%s] %s' % (self.soft.name, message))

    def update_dirs(self):
        self.soft.dirs.update(set([x.replace('${SCRATCH_FOLDER}/', '/')
                                   for x in self.outputs['dirs']]))

    def init_outputs(self):
        self.outputs = {'cmds': {}, 'outs': {}, 'dirs': [], 'bash': [],
                        'io': {('I', 'd'): {}, ('I', 'f'): {},
                               ('O', 'd'): {}, ('O', 'f'): {}}}

    def init_io(self, key):
        if key not in self.soft.io:
            self.soft.io[key] = {}

    def fill_soft_io(self):
        for i, j in itertools.product(*[['I', 'O'], ['d', 'f']]):
            for key, io in self.outputs['io'].get((i, j), {}).items():
                self.init_io((self.sam_pool, key))
                self.soft.io[(self.sam_pool, key)][(i, j)] = io

    def unpack_cmds(self):
        for tech, cmds in self.outputs['cmds'].items():
            self.cmds[(self.sam_pool, tech)] = cmds

    def extract_data(self):
        if self.outputs.get('cmds'):
            self.unpack_cmds()
            self.fill_soft_io()
        if self.soft.name in self.holistics:
            self.soft.outputs = self.outputs['outs']
        else:
            self.soft.outputs[self.sam_pool] = self.outputs['outs']
        self.soft.bash = self.outputs['bash']

    def prep_job(self):
        self.init_outputs()     # initialize data structure collection
        self.call_method()      # collect commands, outputs, io, dirs
        self.extract_data()     # fill the useful self.soft attributes
        self.update_dirs()

    def pooling(self):
        for pool in self.config.pooling_groups:
            self.pools[pool] = {}
            self.soft.outputs[pool] = {}
            pooling(self, pool)

    def call_method(self):
        """Call the command-preparing method from this class (for the
        softwares that are easy to deal with), or from auxillary modules
        located in the softwares submodules path."""
        names = self.soft.name.split('_')
        name = names[0]
        if name in globals():
            globals()[name](self)
        else:
            raise ValueError('No method for software %s' % self.soft.name)

    def register_command(self):
        self.softs[self.soft.name].cmds = dict(self.cmds)
        self.cmds = {}
